from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ...arctrl_core.Table.arc_table import ArcTable
from ...arctrl_core.Table.arc_table_aux import (ColumnValueRefs, ArcTableValues)
from ...arctrl_core.Table.composite_column import CompositeColumn
from ...arctrl_core.Table.composite_header import CompositeHeader
from ...fable_library.array_ import (map as map_1, unzip)
from ...fable_library.list import (reverse, map, FSharpList, fold, singleton, cons, is_empty, head, tail, empty, of_array, to_array, exists, collect as collect_1, sort_by, of_seq, length, iterate_indexed)
from ...fable_library.map_util import add_to_dict
from ...fable_library.range import range_big_int
from ...fable_library.reflection import enum_type
from ...fable_library.seq import (try_find, to_array as to_array_1, delay, map as map_2, collect, singleton as singleton_1, fold as fold_1)
from ...fable_library.string_ import (starts_with_exact, to_fail, printf)
from ...fable_library.types import Array
from ...fable_library.util import (IEnumerable_1, compare_primitives, max)
from ...fs_spreadsheet.Cells.fs_cell import FsCell
from ...fs_spreadsheet.Cells.fs_cells_collection import Dictionary_tryGet
from ...fs_spreadsheet.fs_address import FsAddress
from ...fs_spreadsheet.fs_worksheet import FsWorksheet
from ...fs_spreadsheet.Ranges.fs_range_address import FsRangeAddress
from ...fs_spreadsheet.Tables.fs_table import FsTable
from .composite_column import (from_string_cell_columns, ColumnValueRefs_fromStringCellColumns, fix_deprecated_ioheader, to_string_cell_columns)

__A = TypeVar("__A")

def Aux_List_groupWhen(f: Callable[[__A], bool], list_1: FSharpList[Any]) -> FSharpList[FSharpList[Any]]:
    def mapping(list_3: FSharpList[__A], f: Any=f, list_1: Any=list_1) -> FSharpList[__A]:
        return reverse(list_3)

    def folder(acc: FSharpList[FSharpList[__A]], e: __A, f: Any=f, list_1: Any=list_1) -> FSharpList[FSharpList[__A]]:
        match_value: bool = f(e)
        if match_value:
            return cons(singleton(e), acc)

        elif not is_empty(acc):
            return cons(cons(e, head(acc)), tail(acc))

        else: 
            return singleton(singleton(e))


    return reverse(map(mapping, fold(folder, empty(), list_1)))


def classify_header_order(header: CompositeHeader) -> enum_type("ARCtrl.Spreadsheet.ArcTable.ColumnOrder", int, [("InputClass", 1.0), ("ProtocolClass", 2.0), ("ParamsClass", 3.0), ("OutputClass", 4.0)]):
    if ((((((header.tag == 4) or (header.tag == 5)) or (header.tag == 6)) or (header.tag == 7)) or (header.tag == 8)) or (header.tag == 9)) or (header.tag == 10):
        return 2

    elif (((((header.tag == 0) or (header.tag == 1)) or (header.tag == 2)) or (header.tag == 3)) or (header.tag == 14)) or (header.tag == 13):
        return 3

    elif header.tag == 12:
        return 4

    else: 
        return 1



def classify_column_order(column: CompositeColumn) -> enum_type("ARCtrl.Spreadsheet.ArcTable.ColumnOrder", int, [("InputClass", 1.0), ("ProtocolClass", 2.0), ("ParamsClass", 3.0), ("OutputClass", 4.0)]):
    return classify_header_order(column.Header)


helper_column_strings: FSharpList[str] = of_array(["Term Source REF", "Term Accession Number", "Unit", "Data Format", "Data Selector Format"])

def group_columns_by_header(string_cell_columns: Array[Array[str]]) -> Array[Array[Array[str]]]:
    def mapping(list_4: FSharpList[Array[str]], string_cell_columns: Any=string_cell_columns) -> Array[Array[str]]:
        return to_array(list_4)

    def f(c: Array[str], string_cell_columns: Any=string_cell_columns) -> bool:
        v: str = c[0]
        def predicate(s: str, c: Any=c) -> bool:
            return starts_with_exact(v, s)

        return not exists(predicate, helper_column_strings)

    return map_1(mapping, to_array(Aux_List_groupWhen(f, of_array(string_cell_columns))), None)


def try_annotation_table(sheet: FsWorksheet) -> FsTable | None:
    def predicate(t: FsTable, sheet: Any=sheet) -> bool:
        return starts_with_exact(t.Name, "annotationTable")

    return try_find(predicate, sheet.Tables)


def compose_columns(string_cell_columns: Array[Array[str]]) -> Array[CompositeColumn]:
    def _arrow3584(columns: Array[Array[str]], string_cell_columns: Any=string_cell_columns) -> CompositeColumn:
        return from_string_cell_columns(columns)

    return map_1(_arrow3584, group_columns_by_header(string_cell_columns), None)


def compose_arc_table_values(string_cell_columns: Array[Array[str]]) -> tuple[Array[CompositeHeader], ArcTableValues]:
    value_map: Any = dict([])
    row_count: int = (len(string_cell_columns[0]) - 1) or 0
    def mapping(columns: Array[Array[str]], string_cell_columns: Any=string_cell_columns) -> tuple[CompositeHeader, ColumnValueRefs]:
        return ColumnValueRefs_fromStringCellColumns(value_map, columns)

    pattern_input: tuple[Array[CompositeHeader], Array[ColumnValueRefs]] = unzip(map_1(mapping, group_columns_by_header(string_cell_columns), None))
    return (pattern_input[0], ArcTableValues.from_ref_columns(pattern_input[1], value_map, row_count))


def try_from_fs_worksheet(sheet: FsWorksheet) -> ArcTable | None:
    try: 
        match_value: FsTable | None = try_annotation_table(sheet)
        if match_value is None:
            return None

        else: 
            t: FsTable = match_value
            def _arrow3588(__unit: None=None) -> IEnumerable_1[Array[str]]:
                def _arrow3587(c: int) -> Array[str]:
                    def _arrow3586(__unit: None=None) -> IEnumerable_1[str]:
                        def _arrow3585(r: int) -> IEnumerable_1[str]:
                            match_value_1: FsCell | None = sheet.CellCollection.TryGetCell(r, c)
                            if match_value_1 is None:
                                return singleton_1("")

                            else: 
                                cell: FsCell = match_value_1
                                return singleton_1(cell.ValueAsString())


                        return collect(_arrow3585, range_big_int(1, 1, t.RangeAddress.LastAddress.RowNumber))

                    return to_array_1(delay(_arrow3586))

                return map_2(_arrow3587, range_big_int(1, 1, t.RangeAddress.LastAddress.ColumnNumber))

            pattern_input: tuple[Array[CompositeHeader], ArcTableValues] = compose_arc_table_values(map_1(fix_deprecated_ioheader, to_array_1(delay(_arrow3588)), None))
            return ArcTable.from_arc_table_values(sheet.Name, list(pattern_input[0]), pattern_input[1])


    except Exception as err:
        arg: str = sheet.Name
        arg_1: str = str(err)
        return to_fail(printf("Could not parse table with name \"%s\":\n%s"))(arg)(arg_1)



def to_fs_worksheet(index: int | None, table: ArcTable) -> FsWorksheet:
    string_count: Any = dict([])
    ws: FsWorksheet = FsWorksheet(table.Name)
    if table.ColumnCount == 0:
        return ws

    else: 
        def _arrow3589(column_1: CompositeColumn, index: Any=index, table: Any=table) -> FSharpList[FSharpList[str]]:
            return to_string_cell_columns(column_1)

        def _arrow3590(column: CompositeColumn, index: Any=index, table: Any=table) -> enum_type("ARCtrl.Spreadsheet.ArcTable.ColumnOrder", int, [("InputClass", 1.0), ("ProtocolClass", 2.0), ("ParamsClass", 3.0), ("OutputClass", 4.0)]):
            return classify_column_order(column)

        class ObjectExpr3591:
            @property
            def Compare(self) -> Callable[[enum_type("ARCtrl.Spreadsheet.ArcTable.ColumnOrder", int, [("InputClass", 1.0), ("ProtocolClass", 2.0), ("ParamsClass", 3.0), ("OutputClass", 4.0)]), enum_type("ARCtrl.Spreadsheet.ArcTable.ColumnOrder", int, [("InputClass", 1.0), ("ProtocolClass", 2.0), ("ParamsClass", 3.0), ("OutputClass", 4.0)])], int]:
                return compare_primitives

        columns: FSharpList[FSharpList[str]] = collect_1(_arrow3589, sort_by(_arrow3590, of_seq(table.Columns), ObjectExpr3591()))
        table_row_count: int
        def folder(acc: int, c: FSharpList[str], index: Any=index, table: Any=table) -> int:
            def _arrow3592(x_1: int, y_1: int, acc: Any=acc, c: Any=c) -> int:
                return compare_primitives(x_1, y_1)

            return max(_arrow3592, acc, length(c))

        max_row: int = fold_1(folder, 0, columns) or 0
        table_row_count = 2 if (max_row == 1) else max_row
        table_column_count: int = length(columns) or 0
        name: str = "annotationTable" if (index is None) else (((("" + "annotationTable") + "") + str(index)) + "")
        fs_table: FsTable = ws.Table(name, FsRangeAddress(FsAddress(1, 1), FsAddress(table_row_count, table_column_count)))
        def action_1(col_i: int, col: FSharpList[str], index: Any=index, table: Any=table) -> None:
            def action(row_i: int, string_cell: str, col_i: Any=col_i, col: Any=col) -> None:
                value: str
                if row_i == 0:
                    match_value: str | None = Dictionary_tryGet(string_cell, string_count)
                    if match_value is None:
                        add_to_dict(string_count, string_cell, "")
                        value = string_cell

                    else: 
                        spaces: str = match_value
                        string_count[string_cell] = spaces + " "
                        value = (string_cell + " ") + spaces


                else: 
                    value = string_cell

                address: FsAddress = FsAddress(row_i + 1, col_i + 1)
                fs_table.Cell(address, ws.CellCollection).SetValueAs(value)

            iterate_indexed(action, col)

        iterate_indexed(action_1, columns)
        return ws



__all__ = ["Aux_List_groupWhen", "classify_header_order", "classify_column_order", "helper_column_strings", "group_columns_by_header", "try_annotation_table", "compose_columns", "compose_arc_table_values", "try_from_fs_worksheet", "to_fs_worksheet"]




from __future__ import annotations
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any
from ...arctrl_core.comment import (Remark, Comment)
from ...arctrl_core.Helper.collections_ import Dictionary_init
from ...fable_library.list import (FSharpList, singleton, empty, append, cons, find, exists, initialize as initialize_1, map as map_2)
from ...fable_library.map import (of_seq, try_find)
from ...fable_library.map_util import (get_item_from_dict, add_to_dict)
from ...fable_library.option import (map as map_1, default_arg)
from ...fable_library.reflection import (TypeInfo, string_type, int32_type, tuple_type, class_type, list_type, record_type)
from ...fable_library.seq import (indexed, map, choose, initialize, max_by, try_pick, fold, iterate, is_empty, length as length_1, try_item, delay, append as append_1)
from ...fable_library.string_ import (to_fail, printf)
from ...fable_library.types import (Record, Array)
from ...fable_library.util import (IEnumerable_1, compare_primitives, equals, IEnumerator)
from ...fs_spreadsheet.Cells.fs_cell import (FsCell, DataType)
from ...fs_spreadsheet.DSL.cell_builder import (CellBuilder__AsCellElement_6F87C2ED, CellBuilder__ctor)
from ...fs_spreadsheet.DSL.row_builder import (RowBuilder__ctor, RowBuilder__Combine_19F30600, RowBuilder_get_Empty)
from ...fs_spreadsheet.DSL.types import (SheetEntity_1_some_2B595, Messages_format)
from ...fs_spreadsheet.DSL.row_builder import RowBuilder
from ...fs_spreadsheet.DSL.types import (SheetEntity_1, RowElement, ColumnIndex)
from ...fs_spreadsheet.fs_address import FsAddress
from ...fs_spreadsheet.fs_row import FsRow
from ...fs_spreadsheet.fs_worksheet import FsWorksheet
from ...fs_spreadsheet.Ranges.fs_range_address import FsRangeAddress
from ..collection_aux import (Dictionary_tryGetValue, Seq_trySkip)
from .comment import (Remark__007CRemark_007C__007C, Comment__007CComment_007C__007C, Comment_wrapCommentKey)

def SparseRowModule_fromValues(v: IEnumerable_1[str]) -> IEnumerable_1[tuple[int, str]]:
    return indexed(v)


def SparseRowModule_getValues(i: IEnumerable_1[tuple[int, str]]) -> IEnumerable_1[str]:
    def mapping(tuple: tuple[int, str], i: Any=i) -> str:
        return tuple[1]

    return map(mapping, i)


def SparseRowModule_fromAllValues(v: IEnumerable_1[str | None]) -> IEnumerable_1[tuple[int, str]]:
    def chooser(tupled_arg: tuple[int, str | None], v: Any=v) -> tuple[int, str] | None:
        def _arrow3429(v_1: str, tupled_arg: Any=tupled_arg) -> tuple[int, str]:
            return (tupled_arg[0], v_1)

        return map_1(_arrow3429, tupled_arg[1])

    return choose(chooser, indexed(v))


def SparseRowModule_getAllValues(i: IEnumerable_1[tuple[int, str]]) -> IEnumerable_1[str | None]:
    class ObjectExpr3430:
        @property
        def Compare(self) -> Callable[[int, int], int]:
            return compare_primitives

    m: Any = of_seq(i, ObjectExpr3430())
    def projection(tuple: tuple[int, str], i: Any=i) -> int:
        return tuple[0]

    class ObjectExpr3431:
        @property
        def Compare(self) -> Callable[[int, int], int]:
            return compare_primitives

    def _arrow3432(i_1: int, i: Any=i) -> str | None:
        return try_find(i_1, m)

    return initialize(max_by(projection, i, ObjectExpr3431())[0] + 1, _arrow3432)


def SparseRowModule_fromFsRow(r: FsRow) -> IEnumerable_1[tuple[int, str]]:
    def chooser(c: FsCell, r: Any=r) -> tuple[int, str] | None:
        if equals(c.Value, ""):
            return None

        else: 
            return (c.ColumnNumber - 1, c.ValueAsString())


    return choose(chooser, r.Cells)


def SparseRowModule_tryGetValueAt(i: int, vs: IEnumerable_1[tuple[int, str]]) -> str | None:
    def chooser(tupled_arg: tuple[int, str], i: Any=i, vs: Any=vs) -> str | None:
        if tupled_arg[0] == i:
            return tupled_arg[1]

        else: 
            return None


    return try_pick(chooser, vs)


def SparseRowModule_toDSLRow(vs: IEnumerable_1[tuple[int, str]]) -> FSharpList[RowElement]:
    _builder: RowBuilder = RowBuilder__ctor()
    this_9: SheetEntity_1[FSharpList[RowElement]]
    def f_2(_arg: str | None=None, vs: Any=vs) -> SheetEntity_1[FSharpList[RowElement]]:
        v: str | None = _arg
        if v is None:
            c_1: tuple[tuple[DataType, Any], int | None]
            this_5: SheetEntity_1[tuple[tuple[DataType, Any], int | None]] = CellBuilder__AsCellElement_6F87C2ED(CellBuilder__ctor(), SheetEntity_1_some_2B595(singleton((DataType(0), ""))))
            (pattern_matching_result, errs_1, f_1) = (None, None, None)
            if this_5.tag == 1:
                if equals(this_5.fields[0], empty()):
                    pattern_matching_result = 1

                else: 
                    pattern_matching_result = 2


            elif this_5.tag == 2:
                if equals(this_5.fields[0], empty()):
                    pattern_matching_result = 1

                else: 
                    pattern_matching_result = 2


            else: 
                pattern_matching_result = 0
                errs_1 = this_5.fields[1]
                f_1 = this_5.fields[0]

            if pattern_matching_result == 0:
                c_1 = f_1

            elif pattern_matching_result == 1:
                raise Exception("SheetEntity does not contain Value.")

            elif pattern_matching_result == 2:
                (pattern_matching_result_1, ms_3_1) = (None, None)
                if this_5.tag == 1:
                    pattern_matching_result_1 = 0
                    ms_3_1 = this_5.fields[0]

                elif this_5.tag == 2:
                    pattern_matching_result_1 = 0
                    ms_3_1 = this_5.fields[0]

                else: 
                    pattern_matching_result_1 = 1

                if pattern_matching_result_1 == 0:
                    raise Exception(("SheetEntity does not contain Value: \n\t" + Messages_format(ms_3_1)) + "")

                elif pattern_matching_result_1 == 1:
                    raise Exception("Match failure: FsSpreadsheet.DSL.SheetEntity`1")


            return SheetEntity_1_some_2B595(singleton(RowElement(1, c_1[0]) if (c_1[1] is None) else RowElement(0, ColumnIndex(0, c_1[1]), c_1[0])))

        else: 
            v_1: str = v
            c: tuple[tuple[DataType, Any], int | None]
            this_2: SheetEntity_1[tuple[tuple[DataType, Any], int | None]] = CellBuilder__AsCellElement_6F87C2ED(CellBuilder__ctor(), SheetEntity_1_some_2B595(singleton((DataType(0), v_1))))
            (pattern_matching_result_2, errs, f) = (None, None, None)
            if this_2.tag == 1:
                if equals(this_2.fields[0], empty()):
                    pattern_matching_result_2 = 1

                else: 
                    pattern_matching_result_2 = 2


            elif this_2.tag == 2:
                if equals(this_2.fields[0], empty()):
                    pattern_matching_result_2 = 1

                else: 
                    pattern_matching_result_2 = 2


            else: 
                pattern_matching_result_2 = 0
                errs = this_2.fields[1]
                f = this_2.fields[0]

            if pattern_matching_result_2 == 0:
                c = f

            elif pattern_matching_result_2 == 1:
                raise Exception("SheetEntity does not contain Value.")

            elif pattern_matching_result_2 == 2:
                (pattern_matching_result_3, ms_3) = (None, None)
                if this_2.tag == 1:
                    pattern_matching_result_3 = 0
                    ms_3 = this_2.fields[0]

                elif this_2.tag == 2:
                    pattern_matching_result_3 = 0
                    ms_3 = this_2.fields[0]

                else: 
                    pattern_matching_result_3 = 1

                if pattern_matching_result_3 == 0:
                    raise Exception(("SheetEntity does not contain Value: \n\t" + Messages_format(ms_3)) + "")

                elif pattern_matching_result_3 == 1:
                    raise Exception("Match failure: FsSpreadsheet.DSL.SheetEntity`1")


            return SheetEntity_1_some_2B595(singleton(RowElement(1, c[0]) if (c[1] is None) else RowElement(0, ColumnIndex(0, c[1]), c[0])))


    ns: IEnumerable_1[SheetEntity_1[FSharpList[RowElement]]] = map(f_2, SparseRowModule_getAllValues(vs))
    def folder(state: SheetEntity_1[FSharpList[RowElement]], we: SheetEntity_1[FSharpList[RowElement]], vs: Any=vs) -> SheetEntity_1[FSharpList[RowElement]]:
        return RowBuilder__Combine_19F30600(_builder, state, we)

    this_9 = fold(folder, RowBuilder_get_Empty(), ns)
    (pattern_matching_result_4, errs_2, f_3) = (None, None, None)
    if this_9.tag == 1:
        if equals(this_9.fields[0], empty()):
            pattern_matching_result_4 = 1

        else: 
            pattern_matching_result_4 = 2


    elif this_9.tag == 2:
        if equals(this_9.fields[0], empty()):
            pattern_matching_result_4 = 1

        else: 
            pattern_matching_result_4 = 2


    else: 
        pattern_matching_result_4 = 0
        errs_2 = this_9.fields[1]
        f_3 = this_9.fields[0]

    if pattern_matching_result_4 == 0:
        return f_3

    elif pattern_matching_result_4 == 1:
        raise Exception("SheetEntity does not contain Value.")

    elif pattern_matching_result_4 == 2:
        (pattern_matching_result_5, ms_3_2) = (None, None)
        if this_9.tag == 1:
            pattern_matching_result_5 = 0
            ms_3_2 = this_9.fields[0]

        elif this_9.tag == 2:
            pattern_matching_result_5 = 0
            ms_3_2 = this_9.fields[0]

        else: 
            pattern_matching_result_5 = 1

        if pattern_matching_result_5 == 0:
            raise Exception(("SheetEntity does not contain Value: \n\t" + Messages_format(ms_3_2)) + "")

        elif pattern_matching_result_5 == 1:
            raise Exception("Match failure: FsSpreadsheet.DSL.SheetEntity`1")




def SparseRowModule_readFromSheet(sheet: FsWorksheet) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    def mapping(r: FsRow, sheet: Any=sheet) -> IEnumerable_1[tuple[int, str]]:
        return SparseRowModule_fromFsRow(r)

    return map(mapping, sheet.Rows)


def SparseRowModule_writeToSheet(row_i: int, row: IEnumerable_1[tuple[int, str]], sheet: FsWorksheet) -> None:
    fs_row: FsRow = sheet.RowWithRange(FsRangeAddress(FsAddress(row_i, 1), FsAddress(row_i, 1)), True)
    def action(tupled_arg: tuple[int, str], row_i: Any=row_i, row: Any=row, sheet: Any=sheet) -> None:
        v: str = tupled_arg[1]
        if v.strip() != "":
            fs_row.Item(tupled_arg[0] + 1).SetValueAs(v)


    iterate(action, row)


def _expr3433() -> TypeInfo:
    return record_type("ARCtrl.Spreadsheet.SparseTable", [], SparseTable, lambda: [("Matrix", class_type("System.Collections.Generic.Dictionary`2", [tuple_type(string_type, int32_type), string_type])), ("Keys", list_type(string_type)), ("CommentKeys", list_type(string_type)), ("ColumnCount", int32_type)])


@dataclass(eq = False, repr = False, slots = True)
class SparseTable(Record):
    Matrix: Any
    Keys: FSharpList[str]
    CommentKeys: FSharpList[str]
    ColumnCount: int

SparseTable_reflection = _expr3433

def SparseTable__TryGetValue_11FD62A8(this: SparseTable, key: tuple[str, int]) -> str | None:
    return Dictionary_tryGetValue(key, this.Matrix)


def SparseTable__TryGetValueDefault_5BAE6133(this: SparseTable, default_value: str, key: tuple[str, int]) -> str:
    if key in this.Matrix:
        return get_item_from_dict(this.Matrix, key)

    else: 
        return default_value



def SparseTable_Create_Z2192E64B(matrix: Any | None=None, keys: FSharpList[str] | None=None, comment_keys: FSharpList[str] | None=None, length: int | None=None) -> SparseTable:
    return SparseTable(default_arg(matrix, Dictionary_init()), default_arg(keys, empty()), default_arg(comment_keys, empty()), default_arg(length, 0))


def SparseTable_AddRow(key: str, values: IEnumerable_1[tuple[int, str]], matrix: SparseTable) -> SparseTable:
    def action(tupled_arg: tuple[int, str], key: Any=key, values: Any=values, matrix: Any=matrix) -> None:
        add_to_dict(matrix.Matrix, (key, tupled_arg[0]), tupled_arg[1])

    iterate(action, values)
    def _arrow3434(tuple: tuple[int, str], key: Any=key, values: Any=values, matrix: Any=matrix) -> int:
        return tuple[0]

    class ObjectExpr3435:
        @property
        def Compare(self) -> Callable[[int, int], int]:
            return compare_primitives

    length: int = (0 if is_empty(values) else (1 + max_by(_arrow3434, values, ObjectExpr3435())[0])) or 0
    return SparseTable(matrix.Matrix, append(matrix.Keys, singleton(key)), matrix.CommentKeys, length if (length > matrix.ColumnCount) else matrix.ColumnCount)


def SparseTable_AddEmptyComment(key: str, matrix: SparseTable) -> SparseTable:
    return SparseTable(matrix.Matrix, matrix.Keys, append(matrix.CommentKeys, singleton(key)), matrix.ColumnCount)


def SparseTable_AddComment(key: str, values: IEnumerable_1[tuple[int, str]], matrix: SparseTable) -> SparseTable:
    if length_1(values) == 0:
        return SparseTable_AddEmptyComment(key, matrix)

    else: 
        def action(tupled_arg: tuple[int, str], key: Any=key, values: Any=values, matrix: Any=matrix) -> None:
            add_to_dict(matrix.Matrix, (key, tupled_arg[0]), tupled_arg[1])

        iterate(action, values)
        def _arrow3436(tuple: tuple[int, str], key: Any=key, values: Any=values, matrix: Any=matrix) -> int:
            return tuple[0]

        class ObjectExpr3437:
            @property
            def Compare(self) -> Callable[[int, int], int]:
                return compare_primitives

        length: int = (0 if is_empty(values) else (1 + max_by(_arrow3436, values, ObjectExpr3437())[0])) or 0
        return SparseTable(matrix.Matrix, matrix.Keys, append(matrix.CommentKeys, singleton(key)), length if (length > matrix.ColumnCount) else matrix.ColumnCount)



def SparseTable_FromRows_Z5579EC29(en: IEnumerator[IEnumerable_1[tuple[int, str]]], labels: FSharpList[str], line_number: int, prefix: str | None=None) -> tuple[str | None, int, FSharpList[Remark], SparseTable]:
    try: 
        prefix_1: str = "" if (prefix is None) else (prefix + " ")
        def loop(matrix_mut: SparseTable, remarks_mut: FSharpList[Remark], line_number_1_mut: int) -> tuple[str | None, int, FSharpList[Remark], SparseTable]:
            while True:
                (matrix, remarks, line_number_1) = (matrix_mut, remarks_mut, line_number_1_mut)
                if en.System_Collections_IEnumerator_MoveNext():
                    def mapping(tupled_arg: tuple[int, str], matrix: Any=matrix, remarks: Any=remarks, line_number_1: Any=line_number_1) -> tuple[int, str]:
                        return (tupled_arg[0] - 1, tupled_arg[1])

                    row: IEnumerable_1[tuple[int, str]] = map(mapping, en.System_Collections_Generic_IEnumerator_1_get_Current())
                    def mapping_1(tuple: tuple[int, str], matrix: Any=matrix, remarks: Any=remarks, line_number_1: Any=line_number_1) -> str:
                        return tuple[1]

                    match_value: str | None = map_1(mapping_1, try_item(0, row))
                    vals: IEnumerable_1[tuple[int, str]] | None = Seq_trySkip(1, row)
                    key: str | None = match_value
                    (pattern_matching_result, k, v_1) = (None, None, None)
                    if key is not None:
                        active_pattern_result: str | None = Comment__007CComment_007C__007C(key)
                        if active_pattern_result is not None:
                            if vals is not None:
                                pattern_matching_result = 0
                                k = active_pattern_result
                                v_1 = vals

                            else: 
                                pattern_matching_result = 1


                        else: 
                            pattern_matching_result = 1


                    else: 
                        pattern_matching_result = 1

                    if pattern_matching_result == 0:
                        matrix_mut = SparseTable_AddComment(k, v_1, matrix)
                        remarks_mut = remarks
                        line_number_1_mut = line_number_1 + 1
                        continue

                    elif pattern_matching_result == 1:
                        (pattern_matching_result_1, k_2, k_3, v_3, k_4) = (None, None, None, None, None)
                        active_pattern_result_1: str | None = Remark__007CRemark_007C__007C(key)
                        if active_pattern_result_1 is not None:
                            pattern_matching_result_1 = 0
                            k_2 = active_pattern_result_1

                        elif key is not None:
                            if vals is not None:
                                def _arrow3440(__unit: None=None, matrix: Any=matrix, remarks: Any=remarks, line_number_1: Any=line_number_1) -> bool:
                                    v_2: IEnumerable_1[tuple[int, str]] = vals
                                    k_1: str = key
                                    def _arrow3439(label: str) -> bool:
                                        return k_1 == (prefix_1 + label)

                                    return exists(_arrow3439, labels)

                                if _arrow3440():
                                    pattern_matching_result_1 = 1
                                    k_3 = key
                                    v_3 = vals

                                else: 
                                    pattern_matching_result_1 = 2
                                    k_4 = key


                            else: 
                                pattern_matching_result_1 = 2
                                k_4 = key


                        else: 
                            pattern_matching_result_1 = 3

                        if pattern_matching_result_1 == 0:
                            matrix_mut = matrix
                            remarks_mut = cons(Remark.make(line_number_1, k_2), remarks)
                            line_number_1_mut = line_number_1 + 1
                            continue

                        elif pattern_matching_result_1 == 1:
                            def _arrow3438(label_1: str, matrix: Any=matrix, remarks: Any=remarks, line_number_1: Any=line_number_1) -> bool:
                                return k_3 == (prefix_1 + label_1)

                            matrix_mut = SparseTable_AddRow(find(_arrow3438, labels), v_3, matrix)
                            remarks_mut = remarks
                            line_number_1_mut = line_number_1 + 1
                            continue

                        elif pattern_matching_result_1 == 2:
                            return (k_4, line_number_1, remarks, matrix)

                        elif pattern_matching_result_1 == 3:
                            return (None, line_number_1, remarks, matrix)



                else: 
                    return (None, line_number_1, remarks, matrix)

                break

        return loop(SparseTable_Create_Z2192E64B(), empty(), line_number)

    except Exception as err:
        arg_1: str = str(err)
        return to_fail(printf("Error parsing block in investigation file starting from line number %i: %s"))(line_number)(arg_1)



def SparseTable_ToRows_759CAFC1(matrix: SparseTable, prefix: str | None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    prefix_1: str = "" if (prefix is None) else (prefix + " ")
    def _arrow3446(__unit: None=None, matrix: Any=matrix, prefix: Any=prefix) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
        def _arrow3442(key: str) -> IEnumerable_1[tuple[int, str]]:
            def _arrow3441(i: int) -> str:
                return SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (key, i + 1))

            return SparseRowModule_fromValues(cons(prefix_1 + key, initialize_1(matrix.ColumnCount - 1, _arrow3441)))

        def _arrow3445(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
            def _arrow3444(key_1: str) -> IEnumerable_1[tuple[int, str]]:
                def _arrow3443(i_1: int) -> str:
                    return SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (key_1, i_1 + 1))

                return SparseRowModule_fromValues(cons(Comment_wrapCommentKey(key_1), initialize_1(matrix.ColumnCount - 1, _arrow3443)))

            return map(_arrow3444, matrix.CommentKeys)

        return append_1(map(_arrow3442, matrix.Keys), delay(_arrow3445))

    return delay(_arrow3446)


def SparseTable_GetEmptyComments_3ECCA699(matrix: SparseTable) -> Array[Comment]:
    def mapping(key: str, matrix: Any=matrix) -> Comment:
        return Comment.create(key)

    return list(map_2(mapping, matrix.CommentKeys))


__all__ = ["SparseRowModule_fromValues", "SparseRowModule_getValues", "SparseRowModule_fromAllValues", "SparseRowModule_getAllValues", "SparseRowModule_fromFsRow", "SparseRowModule_tryGetValueAt", "SparseRowModule_toDSLRow", "SparseRowModule_readFromSheet", "SparseRowModule_writeToSheet", "SparseTable_reflection", "SparseTable__TryGetValue_11FD62A8", "SparseTable__TryGetValueDefault_5BAE6133", "SparseTable_Create_Z2192E64B", "SparseTable_AddRow", "SparseTable_AddEmptyComment", "SparseTable_AddComment", "SparseTable_FromRows_Z5579EC29", "SparseTable_ToRows_759CAFC1", "SparseTable_GetEmptyComments_3ECCA699"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...dynamic_obj.dynamic_obj import DynamicObj
from ...dynamic_obj.dyn_obj import (try_get_property_value, set_property, set_optional_property)
from ...fable_library.option import (value as value_1, default_arg)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import (FSharpRef, Array)
from ..ldobject import (LDNode, LDNode_reflection)

def _expr1581() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.ScholarlyArticle", None, ScholarlyArticle, LDNode_reflection())


class ScholarlyArticle(LDNode):
    def __init__(self, id: str, headline: Any=None, identifier: Any=None, additional_type: Array[str] | None=None, author: Any | None=None, url: Any | None=None, creative_work_status: Any | None=None, disambiguating_description: Any | None=None) -> None:
        super().__init__(id, ["schema.org/ScholarlyArticle"], default_arg(additional_type, []))
        this: FSharpRef[ScholarlyArticle] = FSharpRef(None)
        this.contents = self
        self.init_00408: int = 1
        set_property("headline", headline, this.contents)
        set_property("identifier", identifier, this.contents)
        set_optional_property("author", author, this.contents)
        set_optional_property("url", url, this.contents)
        set_optional_property("creativeWorkStatus", creative_work_status, this.contents)
        set_optional_property("disambiguatingDescription", disambiguating_description, this.contents)

    def GetHeadline(self, __unit: None=None) -> str:
        this: ScholarlyArticle = self
        obj: DynamicObj = this
        if try_get_property_value("headline", obj) is not None:
            match_value: str | None
            match_value_1: Any | None = obj.TryGetPropertyValue("headline")
            if match_value_1 is not None:
                o: Any = value_1(match_value_1)
                match_value = o if (str(type(o)) == "<class \'str\'>") else None

            else: 
                match_value = None

            if match_value is None:
                raise Exception(((((("Property \'" + "headline") + "\' is set on this \'") + "ScholarlyArticle") + "\' object but cannot be cast to \'") + "String") + "\'")

            else: 
                return match_value


        else: 
            raise Exception(((("No property \'" + "headline") + "\' set on this \'") + "ScholarlyArticle") + "\' object although it is mandatory. Was it created correctly?")


    @staticmethod
    def get_headline() -> Callable[[ScholarlyArticle], str]:
        def _arrow1579(s: ScholarlyArticle) -> str:
            return s.GetHeadline()

        return _arrow1579

    def GetIdentifier(self, __unit: None=None) -> str:
        this: ScholarlyArticle = self
        obj: DynamicObj = this
        if try_get_property_value("identifier", obj) is not None:
            match_value: str | None
            match_value_1: Any | None = obj.TryGetPropertyValue("identifier")
            if match_value_1 is not None:
                o: Any = value_1(match_value_1)
                match_value = o if (str(type(o)) == "<class \'str\'>") else None

            else: 
                match_value = None

            if match_value is None:
                raise Exception(((((("Property \'" + "identifier") + "\' is set on this \'") + "ScholarlyArticle") + "\' object but cannot be cast to \'") + "String") + "\'")

            else: 
                return match_value


        else: 
            raise Exception(((("No property \'" + "identifier") + "\' set on this \'") + "ScholarlyArticle") + "\' object although it is mandatory. Was it created correctly?")


    @staticmethod
    def get_identifier() -> Callable[[ScholarlyArticle], str]:
        def _arrow1580(s: ScholarlyArticle) -> str:
            return s.GetIdentifier()

        return _arrow1580


ScholarlyArticle_reflection = _expr1581

def ScholarlyArticle__ctor_Z22702026(id: str, headline: Any=None, identifier: Any=None, additional_type: Array[str] | None=None, author: Any | None=None, url: Any | None=None, creative_work_status: Any | None=None, disambiguating_description: Any | None=None) -> ScholarlyArticle:
    return ScholarlyArticle(id, headline, identifier, additional_type, author, url, creative_work_status, disambiguating_description)


__all__ = ["ScholarlyArticle_reflection"]




from __future__ import annotations
from typing import Any
from ...arctrl_core.arc_types import ArcStudy
from ...arctrl_core.comment import Comment
from ...arctrl_core.data_map import DataMap
from ...arctrl_core.Helper.collections_ import (Option_fromSeq, ResizeArray_map, Option_fromValueWithDefault)
from ...arctrl_core.person import Person
from ...arctrl_core.publication import Publication
from ...arctrl_core.Table.arc_tables import ArcTables
from ...arctrl_file_system.file_system import FileSystem
from ...arctrl_rocrate.ldcontext import LDContext
from ...arctrl_rocrate.ldobject import (LDNode, LDGraph)
from ...arctrl_rocrate.LDTypes.dataset import LDDataset
from ...fable_library.date import now
from ...fable_library.list import of_seq
from ...fable_library.option import (bind, map)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import Array
from .assay import AssayConversion_getDataFilesFromProcesses_6BABD1B0
from .basic import (BaseTypes_composeComment_Z13201A7E, BaseTypes_decomposeComment_Z2F770004)
from .datamap import (DatamapConversion_composeFragmentDescriptors_Z8923FA3, DatamapConversion_decomposeFragmentDescriptors_Z6E59645F)
from .date_time import (try_from_string, to_string)
from .person import (PersonConversion_composePerson_Z64D846DC, PersonConversion_decomposePerson_Z6839B9E8)
from .scholarly_article import (ScholarlyArticleConversion_composeScholarlyArticle_D324A6D, ScholarlyArticleConversion_decomposeScholarlyArticle_Z6839B9E8)
from .table import (ARCtrl_ArcTables__ArcTables_GetProcesses_5E660E5C, ARCtrl_ArcTables__ArcTables_fromProcesses_Static_Z27F0B586)

def _expr3972() -> TypeInfo:
    return class_type("ARCtrl.Conversion.StudyConversion", None, StudyConversion)


class StudyConversion:
    ...

StudyConversion_reflection = _expr3972

def StudyConversion_composeStudy_ZFE0E38E(study: ArcStudy, fs: FileSystem | None=None) -> LDNode:
    def _arrow3973(s: str, study: Any=study, fs: Any=fs) -> Any | None:
        return try_from_string(s)

    date_created: Any | None = bind(_arrow3973, study.SubmissionDate)
    def _arrow3974(s_1: str, study: Any=study, fs: Any=fs) -> Any | None:
        return try_from_string(s_1)

    date_published: Any | None = bind(_arrow3974, study.PublicReleaseDate)
    date_modified: Any = now()
    def f(p: Publication, study: Any=study, fs: Any=fs) -> LDNode:
        return ScholarlyArticleConversion_composeScholarlyArticle_D324A6D(p)

    publications: Array[LDNode] | None = Option_fromSeq(ResizeArray_map(f, study.Publications))
    def f_1(c: Person, study: Any=study, fs: Any=fs) -> LDNode:
        return PersonConversion_composePerson_Z64D846DC(c)

    creators: Array[LDNode] | None = Option_fromSeq(ResizeArray_map(f_1, study.Contacts))
    process_sequence: Array[LDNode] | None = Option_fromSeq(list(ARCtrl_ArcTables__ArcTables_GetProcesses_5E660E5C(ArcTables(study.Tables), None, study.Identifier, fs)))
    def mapping(datamap: DataMap, study: Any=study, fs: Any=fs) -> Array[LDNode]:
        return DatamapConversion_composeFragmentDescriptors_Z8923FA3(datamap)

    fragment_descriptors: Array[LDNode] | None = map(mapping, study.DataMap)
    def mapping_1(ps: Array[LDNode], study: Any=study, fs: Any=fs) -> Array[LDNode]:
        return AssayConversion_getDataFilesFromProcesses_6BABD1B0(ps, fragment_descriptors)

    data_files: Array[LDNode] | None = map(mapping_1, process_sequence)
    def f_2(c_1: Comment, study: Any=study, fs: Any=fs) -> LDNode:
        return BaseTypes_composeComment_Z13201A7E(c_1)

    comments: Array[LDNode] | None = Option_fromSeq(ResizeArray_map(f_2, study.Comments))
    return LDDataset.create_study(study.Identifier, None, creators, date_created, date_published, date_modified, study.Description, data_files, study.Title, publications, fragment_descriptors, comments, None, process_sequence)


def StudyConversion_decomposeStudy_Z6839B9E8(study: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> ArcStudy:
    def _arrow3975(d: Any, study: Any=study, graph: Any=graph, context: Any=context) -> str:
        return to_string(d)

    date_created: str | None = map(_arrow3975, LDDataset.try_get_date_created_as_date_time(study, context))
    def _arrow3976(d_1: Any, study: Any=study, graph: Any=graph, context: Any=context) -> str:
        return to_string(d_1)

    date_published: str | None = map(_arrow3976, LDDataset.try_get_date_published_as_date_time(study, context))
    def f(p: LDNode, study: Any=study, graph: Any=graph, context: Any=context) -> Publication:
        return ScholarlyArticleConversion_decomposeScholarlyArticle_Z6839B9E8(p, graph, context)

    publications: Array[Publication] = ResizeArray_map(f, LDDataset.get_citations(study, graph, context))
    def f_1(c: LDNode, study: Any=study, graph: Any=graph, context: Any=context) -> Person:
        return PersonConversion_decomposePerson_Z6839B9E8(c, graph, context)

    creators: Array[Person] = ResizeArray_map(f_1, LDDataset.get_creators(study, graph, context))
    data_map: DataMap | None
    v: DataMap = DatamapConversion_decomposeFragmentDescriptors_Z6E59645F(LDDataset.get_variable_measured_as_fragment_descriptors(study, graph, context), graph, context)
    data_map = Option_fromValueWithDefault(DataMap.init(), v)
    tables: ArcTables = ARCtrl_ArcTables__ArcTables_fromProcesses_Static_Z27F0B586(of_seq(LDDataset.get_abouts_as_lab_process(study, graph, context)), graph, context)
    def f_2(c_1: LDNode, study: Any=study, graph: Any=graph, context: Any=context) -> Comment:
        return BaseTypes_decomposeComment_Z2F770004(c_1, context)

    comments: Array[Comment] = ResizeArray_map(f_2, LDDataset.get_comments(study, graph, context))
    return ArcStudy.create(LDDataset.get_identifier_as_string(study, context), LDDataset.try_get_name_as_string(study, context), LDDataset.try_get_description_as_string(study, context), date_created, date_published, publications, creators, None, tables.Tables, data_map, None, comments)


__all__ = ["StudyConversion_reflection", "StudyConversion_composeStudy_ZFE0E38E", "StudyConversion_decomposeStudy_Z6839B9E8"]




from __future__ import annotations
from typing import Any
from ...arctrl_core.data_context import DataContext
from ...arctrl_core.data_map import DataMap
from ...arctrl_core.Helper.collections_ import ResizeArray_map
from ...arctrl_rocrate.ldcontext import LDContext
from ...arctrl_rocrate.ldobject import (LDNode, LDGraph)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import Array
from .basic import (BaseTypes_composeFragmentDescriptor_Z4C0BEF62, BaseTypes_decomposeFragmentDescriptor_Z6839B9E8)

def _expr3877() -> TypeInfo:
    return class_type("ARCtrl.Conversion.DatamapConversion", None, DatamapConversion)


class DatamapConversion:
    ...

DatamapConversion_reflection = _expr3877

def DatamapConversion_composeFragmentDescriptors_Z8923FA3(datamap: DataMap) -> Array[LDNode]:
    def f(dc: DataContext, datamap: Any=datamap) -> LDNode:
        return BaseTypes_composeFragmentDescriptor_Z4C0BEF62(dc)

    return ResizeArray_map(f, datamap.DataContexts)


def DatamapConversion_decomposeFragmentDescriptors_Z6E59645F(fragment_descriptors: Array[LDNode], graph: LDGraph | None=None, context: LDContext | None=None) -> DataMap:
    def f(fd: LDNode, fragment_descriptors: Any=fragment_descriptors, graph: Any=graph, context: Any=context) -> DataContext:
        return BaseTypes_decomposeFragmentDescriptor_Z6839B9E8(fd, graph, context)

    return DataMap(ResizeArray_map(f, fragment_descriptors))


__all__ = ["DatamapConversion_reflection", "DatamapConversion_composeFragmentDescriptors_Z8923FA3", "DatamapConversion_decomposeFragmentDescriptors_Z6E59645F"]




from __future__ import annotations
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any
from ...fable_library.option import (map, bind, default_arg)
from ...fable_library.reflection import (TypeInfo, string_type, option_type, record_type)
from ...fable_library.string_ import (to_text, printf)
from ...fable_library.types import (to_string, Record)
from ...fable_library.util import int32_to_string
from ..ontology_annotation import (OntologyAnnotation, OntologyAnnotation_reflection)
from ..value import (Value as Value_1, Value_reflection)
from .material_attribute import (MaterialAttribute__get_NameText, MaterialAttribute, MaterialAttribute_reflection, MaterialAttribute__get_TryNameText, MaterialAttribute__MapCategory_658CFBF6, MaterialAttribute_create_A220A8A, MaterialAttribute__SetCategory_ZDED3A0F)

def _expr756() -> TypeInfo:
    return record_type("ARCtrl.Process.MaterialAttributeValue", [], MaterialAttributeValue, lambda: [("ID", option_type(string_type)), ("Category", option_type(MaterialAttribute_reflection())), ("Value", option_type(Value_reflection())), ("Unit", option_type(OntologyAnnotation_reflection()))])


@dataclass(eq = False, repr = False, slots = True)
class MaterialAttributeValue(Record):
    ID: str | None
    Category: MaterialAttribute | None
    Value: Value_1 | None
    Unit: OntologyAnnotation | None
    def Print(self, __unit: None=None) -> str:
        this: MaterialAttributeValue = self
        return to_string(this)

    def PrintCompact(self, __unit: None=None) -> str:
        this: MaterialAttributeValue = self
        category: str | None = map(MaterialAttribute__get_NameText, this.Category)
        def mapping_1(oa: OntologyAnnotation) -> str:
            return oa.NameText

        unit: str | None = map(mapping_1, this.Unit)
        def mapping_2(v: Value_1) -> str:
            s: str = v.PrintCompact()
            if unit is None:
                return s

            else: 
                return (s + " ") + unit


        value: str | None = map(mapping_2, this.Value)
        def _arrow753(__unit: None=None) -> str:
            value_2: str = value
            return value_2

        def _arrow754(__unit: None=None) -> str:
            category_2: str = category
            return (category_2 + ":") + "No Value"

        def _arrow755(__unit: None=None) -> str:
            category_1: str = category
            value_1: str = value
            return (category_1 + ":") + value_1

        return ("" if (value is None) else _arrow753()) if (category is None) else (_arrow754() if (value is None) else _arrow755())

    def AlternateName(self, __unit: None=None) -> str | None:
        return None

    def MeasurementMethod(self, __unit: None=None) -> str | None:
        return None

    def Description(self, __unit: None=None) -> str | None:
        return None

    def GetCategory(self, __unit: None=None) -> OntologyAnnotation | None:
        this: MaterialAttributeValue = self
        def binder(x: MaterialAttribute) -> OntologyAnnotation | None:
            return x.CharacteristicType

        return bind(binder, this.Category)

    def GetValue(self, __unit: None=None) -> Value_1 | None:
        this: MaterialAttributeValue = self
        return this.Value

    def GetUnit(self, __unit: None=None) -> OntologyAnnotation | None:
        this: MaterialAttributeValue = self
        return this.Unit

    def GetAdditionalType(self, __unit: None=None) -> str:
        return "MaterialAttributeValue"


MaterialAttributeValue_reflection = _expr756

def MaterialAttributeValue_make(id: str | None=None, category: MaterialAttribute | None=None, value: Value_1 | None=None, unit: OntologyAnnotation | None=None) -> MaterialAttributeValue:
    return MaterialAttributeValue(id, category, value, unit)


def MaterialAttributeValue_create_ZE1D108D(Id: str | None=None, Category: MaterialAttribute | None=None, Value: Value_1 | None=None, Unit: OntologyAnnotation | None=None) -> MaterialAttributeValue:
    return MaterialAttributeValue_make(Id, Category, Value, Unit)


def MaterialAttributeValue_get_empty(__unit: None=None) -> MaterialAttributeValue:
    return MaterialAttributeValue_create_ZE1D108D()


def MaterialAttributeValue__get_NameText(this: MaterialAttributeValue) -> str:
    def mapping(oa: MaterialAttribute, this: Any=this) -> str:
        return MaterialAttribute__get_NameText(oa)

    return default_arg(map(mapping, this.Category), "")


def MaterialAttributeValue__get_TryNameText(this: MaterialAttributeValue) -> str | None:
    def binder(oa: MaterialAttribute, this: Any=this) -> str | None:
        return MaterialAttribute__get_TryNameText(oa)

    return bind(binder, this.Category)


def MaterialAttributeValue__get_ValueText(this: MaterialAttributeValue) -> str:
    def mapping(oa: Value_1, this: Any=this) -> str:
        if oa.tag == 2:
            return to_string(oa.fields[0])

        elif oa.tag == 1:
            return int32_to_string(oa.fields[0])

        elif oa.tag == 3:
            return oa.fields[0]

        else: 
            return oa.fields[0].NameText


    return default_arg(map(mapping, this.Value), "")


def MaterialAttributeValue__get_ValueWithUnitText(this: MaterialAttributeValue) -> str:
    def mapping(oa: OntologyAnnotation, this: Any=this) -> str:
        return oa.NameText

    unit: str | None = map(mapping, this.Unit)
    v: str = MaterialAttributeValue__get_ValueText(this)
    if unit is None:
        return v

    else: 
        u: str = unit
        return to_text(printf("%s %s"))(v)(u)



def MaterialAttributeValue__MapCategory_658CFBF6(this: MaterialAttributeValue, f: Callable[[OntologyAnnotation], OntologyAnnotation]) -> MaterialAttributeValue:
    def mapping(p: MaterialAttribute, this: Any=this, f: Any=f) -> MaterialAttribute:
        return MaterialAttribute__MapCategory_658CFBF6(p, f)

    return MaterialAttributeValue(this.ID, map(mapping, this.Category), this.Value, this.Unit)


def MaterialAttributeValue__SetCategory_ZDED3A0F(this: MaterialAttributeValue, c: OntologyAnnotation) -> MaterialAttributeValue:
    def _arrow757(__unit: None=None, this: Any=this, c: Any=c) -> MaterialAttribute | None:
        match_value: MaterialAttribute | None = this.Category
        return MaterialAttribute_create_A220A8A(None, c) if (match_value is None) else MaterialAttribute__SetCategory_ZDED3A0F(match_value, c)

    return MaterialAttributeValue(this.ID, _arrow757(), this.Value, this.Unit)


def MaterialAttributeValue_tryGetNameText_Z772273B8(mv: MaterialAttributeValue) -> str | None:
    return MaterialAttributeValue__get_TryNameText(mv)


def MaterialAttributeValue_getNameAsString_Z772273B8(mv: MaterialAttributeValue) -> str | None:
    return MaterialAttributeValue__get_TryNameText(mv)


def MaterialAttributeValue_nameEqualsString(name: str, mv: MaterialAttributeValue) -> bool:
    return MaterialAttributeValue__get_NameText(mv) == name


def MaterialAttributeValue_createAsPV(alternate_name: str | None=None, measurement_method: str | None=None, description: str | None=None, category: OntologyAnnotation | None=None, value: Value_1 | None=None, unit: OntologyAnnotation | None=None) -> MaterialAttributeValue:
    def mapping(c: OntologyAnnotation, alternate_name: Any=alternate_name, measurement_method: Any=measurement_method, description: Any=description, category: Any=category, value: Any=value, unit: Any=unit) -> MaterialAttribute:
        return MaterialAttribute_create_A220A8A(None, c)

    return MaterialAttributeValue_create_ZE1D108D(None, map(mapping, category), value, unit)


__all__ = ["MaterialAttributeValue_reflection", "MaterialAttributeValue_make", "MaterialAttributeValue_create_ZE1D108D", "MaterialAttributeValue_get_empty", "MaterialAttributeValue__get_NameText", "MaterialAttributeValue__get_TryNameText", "MaterialAttributeValue__get_ValueText", "MaterialAttributeValue__get_ValueWithUnitText", "MaterialAttributeValue__MapCategory_658CFBF6", "MaterialAttributeValue__SetCategory_ZDED3A0F", "MaterialAttributeValue_tryGetNameText_Z772273B8", "MaterialAttributeValue_getNameAsString_Z772273B8", "MaterialAttributeValue_nameEqualsString", "MaterialAttributeValue_createAsPV"]




from __future__ import annotations
from typing import Any
from ..fable_library.int32 import parse
from ..fable_library.reg_exp import (match, create, get_item, groups)
from .regex import (KeyPattern, ValuePattern, LineCommentPattern, InlineSequencePattern, SequenceOpenerPattern, SequenceCloserPattern, InlineJSONPattern, JSONOpenerPattern, JSONCloserPattern)
from .yamlicious_types import PreprocessorElement

def _007CKey_007C__007C(input: PreprocessorElement) -> dict[str, Any] | None:
    if input.tag == 2:
        m: Any = match(create(KeyPattern), input.fields[0])
        if m is not None:
            def _arrow399(__unit: None=None, input: Any=input) -> int | None:
                v: str = get_item(groups(m), "comment") or ""
                return None if (v == "") else parse(v, 511, False, 32)

            return {
                "Comment": _arrow399(),
                "Key": get_item(groups(m), "key") or ""
            }

        else: 
            return None


    else: 
        return None



def _007CKeyValue_007C__007C(input: PreprocessorElement) -> dict[str, Any] | None:
    if input.tag == 2:
        m: Any = match(create("^(?P<key>[^\\{{\\[]+):\\s+(?P<value>.*)$"), input.fields[0])
        if m is not None:
            v: str = get_item(groups(m), "value") or "".strip()
            return {
                "Key": get_item(groups(m), "key") or "",
                "Value": v
            }

        else: 
            return None


    else: 
        return None



def _007CYamlValue_007C__007C(input: PreprocessorElement) -> dict[str, Any] | None:
    if input.tag == 2:
        m: Any = match(create(ValuePattern), input.fields[0])
        if m is not None:
            def _arrow400(__unit: None=None, input: Any=input) -> int | None:
                v: str = get_item(groups(m), "comment") or ""
                return None if (v == "") else parse(v, 511, False, 32)

            return {
                "Comment": _arrow400(),
                "Value": get_item(groups(m), "value") or "".strip()
            }

        else: 
            return None


    else: 
        return None



def _007CYamlComment_007C__007C(input: PreprocessorElement) -> dict[str, Any] | None:
    if input.tag == 2:
        m: Any = match(create(LineCommentPattern), input.fields[0])
        if m is not None:
            return {
                "Comment": parse(get_item(groups(m), "comment") or "", 511, False, 32)
            }

        else: 
            return None


    else: 
        return None



def _007CSequenceMinusOpener_007C__007C(input: PreprocessorElement) -> dict[str, Any] | None:
    if input.tag == 2:
        m: Any = match(create("^-(\\s+(?P<value>.*))?$"), input.fields[0])
        if m is not None:
            def _arrow401(__unit: None=None, input: Any=input) -> str | None:
                v: str = get_item(groups(m), "value") or "".strip()
                return None if (v == "") else v

            return {
                "Value": _arrow401()
            }

        else: 
            return None


    else: 
        return None



def _007CInlineSequence_007C__007C(input: PreprocessorElement) -> dict[str, Any] | None:
    if input.tag == 2:
        m: Any = match(create(InlineSequencePattern), input.fields[0])
        if m is not None:
            def _arrow402(__unit: None=None, input: Any=input) -> int | None:
                v: str = get_item(groups(m), "comment") or ""
                return None if (v == "") else parse(v, 511, False, 32)

            return {
                "Comment": _arrow402(),
                "Value": get_item(groups(m), "inlineSequence") or ""
            }

        else: 
            return None


    else: 
        return None



def _007CSequenceSquareOpener_007C__007C(input: PreprocessorElement) -> dict[str, Any] | None:
    if input.tag == 2:
        m: Any = match(create(SequenceOpenerPattern), input.fields[0])
        if m is not None:
            def _arrow403(__unit: None=None, input: Any=input) -> int | None:
                v: str = get_item(groups(m), "comment") or ""
                return None if (v == "") else parse(v, 511, False, 32)

            return {
                "Comment": _arrow403()
            }

        else: 
            return None


    else: 
        return None



def _007CSequenceSquareCloser_007C__007C(input: PreprocessorElement) -> dict[str, Any] | None:
    if input.tag == 2:
        m: Any = match(create(SequenceCloserPattern), input.fields[0])
        if m is not None:
            def _arrow404(__unit: None=None, input: Any=input) -> int | None:
                v: str = get_item(groups(m), "comment") or ""
                return None if (v == "") else parse(v, 511, False, 32)

            return {
                "Comment": _arrow404()
            }

        else: 
            return None


    else: 
        return None



def _007CInlineJSON_007C__007C(input: PreprocessorElement) -> dict[str, Any] | None:
    if input.tag == 2:
        m: Any = match(create(InlineJSONPattern), input.fields[0])
        if m is not None:
            def _arrow405(__unit: None=None, input: Any=input) -> int | None:
                v: str = get_item(groups(m), "comment") or ""
                return None if (v == "") else parse(v, 511, False, 32)

            return {
                "Comment": _arrow405(),
                "Value": get_item(groups(m), "inlineSequence") or ""
            }

        else: 
            return None


    else: 
        return None



def _007CJSONKeyOpener_007C__007C(input: PreprocessorElement) -> dict[str, Any] | None:
    if input.tag == 2:
        m: Any = match(create(JSONOpenerPattern), input.fields[0])
        if m is not None:
            def _arrow406(__unit: None=None, input: Any=input) -> int | None:
                v: str = get_item(groups(m), "comment") or ""
                return None if (v == "") else parse(v, 511, False, 32)

            return {
                "Comment": _arrow406(),
                "Key": get_item(groups(m), "key") or ""
            }

        else: 
            return None


    else: 
        return None



def _007CJSONCloser_007C__007C(input: PreprocessorElement) -> dict[str, Any] | None:
    if input.tag == 2:
        m: Any = match(create(JSONCloserPattern), input.fields[0])
        if m is not None:
            def _arrow407(__unit: None=None, input: Any=input) -> int | None:
                v: str = get_item(groups(m), "comment") or ""
                return None if (v == "") else parse(v, 511, False, 32)

            return {
                "Comment": _arrow407()
            }

        else: 
            return None


    else: 
        return None



def _007CSchemaNamespace_007C__007C(input: PreprocessorElement) -> dict[str, Any] | None:
    if input.tag == 2:
        m: Any = match(create("^\\$(?P<key>[a-zA-Z0-9\\s:]+):$"), input.fields[0])
        if m is not None:
            return {
                "Key": get_item(groups(m), "key") or ""
            }

        else: 
            return None


    else: 
        return None



__all__ = ["_007CKey_007C__007C", "_007CKeyValue_007C__007C", "_007CYamlValue_007C__007C", "_007CYamlComment_007C__007C", "_007CSequenceMinusOpener_007C__007C", "_007CInlineSequence_007C__007C", "_007CSequenceSquareOpener_007C__007C", "_007CSequenceSquareCloser_007C__007C", "_007CInlineJSON_007C__007C", "_007CJSONKeyOpener_007C__007C", "_007CJSONCloser_007C__007C", "_007CSchemaNamespace_007C__007C"]




from __future__ import annotations
from typing import Any
from ...arctrl_core.comment import Comment
from ...arctrl_core.Helper.collections_ import (ResizeArray_map, Option_fromSeq, ResizeArray_tryPick)
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.publication import Publication
from ...arctrl_json.ROCrate.ldnode import (decoder as decoder_1, encoder)
from ...arctrl_rocrate.ldcontext import LDContext
from ...arctrl_rocrate.ldobject import (LDNode, LDGraph)
from ...arctrl_rocrate.LDTypes.person import LDPerson
from ...arctrl_rocrate.LDTypes.property_value import LDPropertyValue
from ...arctrl_rocrate.LDTypes.scholarly_article import LDScholarlyArticle
from ...fable_library.option import (map as map_1, value as value_4)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.result import FSharpResult_2
from ...fable_library.seq import (map, is_empty, filter, to_list, delay, append, singleton, empty)
from ...fable_library.string_ import (to_text, printf, join)
from ...fable_library.system_text import (StringBuilder__ctor, StringBuilder__Append_244C7CD6, StringBuilder__Clear)
from ...fable_library.types import (Array, to_string)
from ...fable_library.util import (ignore, get_enumerator, IEnumerable_1)
from ...thoth_json_python.decode import Decode_fromString
from ...thoth_json_python.encode import to_string as to_string_1
from .basic import (BaseTypes_composeComment_Z13201A7E, BaseTypes_composeDefinedTerm_ZDED3A0F, BaseTypes_decomposeComment_Z2F770004, BaseTypes_decomposeDefinedTerm_Z2F770004)

def _expr3954() -> TypeInfo:
    return class_type("ARCtrl.Conversion.ScholarlyArticleConversion", None, ScholarlyArticleConversion)


class ScholarlyArticleConversion:
    ...

ScholarlyArticleConversion_reflection = _expr3954

def ScholarlyArticleConversion_composeAuthor_Z721C83C5(author: str) -> LDNode:
    try: 
        match_value: FSharpResult_2[LDNode, str] = Decode_fromString(decoder_1, author)
        if match_value.tag == 1:
            raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

        else: 
            return match_value.fields[0]


    except Exception as match_value_1:
        return LDPerson.create(author)



def ScholarlyArticleConversion_splitAuthors_Z721C83C5(a: str) -> Array[str]:
    bracket_count: int = 0
    authors: Array[str] = []
    sb: Any = StringBuilder__ctor()
    with get_enumerator(list(a)) as enumerator:
        while enumerator.System_Collections_IEnumerator_MoveNext():
            c: str = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
            if c == "{":
                bracket_count = (bracket_count + 1) or 0
                ignore(StringBuilder__Append_244C7CD6(sb, c))

            elif c == "}":
                bracket_count = (bracket_count - 1) or 0
                ignore(StringBuilder__Append_244C7CD6(sb, c))

            elif (bracket_count == 0) if (c == ",") else False:
                (authors.append(to_string(sb)))
                ignore(StringBuilder__Clear(sb))

            else: 
                ignore(StringBuilder__Append_244C7CD6(sb, c))

    (authors.append(to_string(sb)))
    return authors


def ScholarlyArticleConversion_composeAuthors_Z721C83C5(authors: str) -> Array[LDNode]:
    def mapping(author: str, authors: Any=authors) -> LDNode:
        return ScholarlyArticleConversion_composeAuthor_Z721C83C5(author)

    return list(map(mapping, ScholarlyArticleConversion_splitAuthors_Z721C83C5(authors)))


def ScholarlyArticleConversion_decomposeAuthor_Z2F770004(author: LDNode, context: LDContext | None=None) -> str:
    def predicate(n: str, author: Any=author, context: Any=context) -> bool:
        return n != LDPerson.given_name()

    if is_empty(filter(predicate, author.GetPropertyNames(context))):
        return LDPerson.get_given_name_as_string(author, context)

    else: 
        return to_string_1(0, encoder(author))



def ScholarlyArticleConversion_decomposeAuthors_1AAAE9A5(authors: Array[LDNode], context: LDContext | None=None) -> str:
    def f(a: LDNode, authors: Any=authors, context: Any=context) -> str:
        return ScholarlyArticleConversion_decomposeAuthor_Z2F770004(a, context)

    return join(",", ResizeArray_map(f, authors))


def ScholarlyArticleConversion_composeScholarlyArticle_D324A6D(publication: Publication) -> LDNode:
    title: str
    match_value: str | None = publication.Title
    if match_value is None:
        raise Exception("Publication must have a title")

    else: 
        title = match_value

    def mapping(authors: str, publication: Any=publication) -> Array[LDNode]:
        return ScholarlyArticleConversion_composeAuthors_Z721C83C5(authors)

    authors_1: Array[LDNode] | None = map_1(mapping, publication.Authors)
    def f(comment: Comment, publication: Any=publication) -> LDNode:
        return BaseTypes_composeComment_Z13201A7E(comment)

    comments: Array[LDNode] | None = Option_fromSeq(ResizeArray_map(f, publication.Comments))
    def _arrow3956(__unit: None=None, publication: Any=publication) -> IEnumerable_1[LDNode]:
        def _arrow3955(__unit: None=None) -> IEnumerable_1[LDNode]:
            return singleton(LDPropertyValue.create_pub_med_id(value_4(publication.PubMedID))) if ((value_4(publication.PubMedID) != "") if (publication.PubMedID is not None) else False) else empty()

        return append(singleton(LDPropertyValue.create_doi(value_4(publication.DOI))) if ((value_4(publication.DOI) != "") if (publication.DOI is not None) else False) else empty(), delay(_arrow3955))

    identifiers: Array[LDNode] = list(to_list(delay(_arrow3956)))
    def mapping_1(term: OntologyAnnotation, publication: Any=publication) -> LDNode:
        return BaseTypes_composeDefinedTerm_ZDED3A0F(term)

    status: LDNode | None = map_1(mapping_1, publication.Status)
    return LDScholarlyArticle.create(title, identifiers, None, authors_1, None, status, comments)


def ScholarlyArticleConversion_decomposeScholarlyArticle_Z6839B9E8(sa: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Publication:
    title: str = LDScholarlyArticle.get_headline_as_string(sa, context)
    def mapping(a: Array[LDNode], sa: Any=sa, graph: Any=graph, context: Any=context) -> str:
        return ScholarlyArticleConversion_decomposeAuthors_1AAAE9A5(a, context)

    authors: str | None = map_1(mapping, Option_fromSeq(LDScholarlyArticle.get_authors(sa, graph, context)))
    def f(c: LDNode, sa: Any=sa, graph: Any=graph, context: Any=context) -> Comment:
        return BaseTypes_decomposeComment_Z2F770004(c, context)

    comments: Array[Comment] = ResizeArray_map(f, LDScholarlyArticle.get_comments(sa, graph, context))
    def mapping_1(s: LDNode, sa: Any=sa, graph: Any=graph, context: Any=context) -> OntologyAnnotation:
        return BaseTypes_decomposeDefinedTerm_Z2F770004(s, context)

    status: OntologyAnnotation | None = map_1(mapping_1, LDScholarlyArticle.try_get_creative_work_status(sa, graph, context))
    identifiers: Array[LDNode] = LDScholarlyArticle.get_identifiers_as_property_value(sa, graph, context)
    doi: str | None
    def f_1(i: LDNode, sa: Any=sa, graph: Any=graph, context: Any=context) -> str | None:
        return LDPropertyValue.try_get_as_doi(i, context)

    _arg: str | None = ResizeArray_tryPick(f_1, identifiers)
    doi = LDScholarlyArticle.try_get_same_as_as_string(sa, context) if (_arg is None) else _arg
    pub_med_id: str | None
    def f_2(i_1: LDNode, sa: Any=sa, graph: Any=graph, context: Any=context) -> str | None:
        return LDPropertyValue.try_get_as_pub_med_id(i_1, context)

    _arg_1: str | None = ResizeArray_tryPick(f_2, identifiers)
    pub_med_id = LDScholarlyArticle.try_get_url_as_string(sa, context) if (_arg_1 is None) else _arg_1
    return Publication.create(pub_med_id, doi, authors, title, status, comments)


__all__ = ["ScholarlyArticleConversion_reflection", "ScholarlyArticleConversion_composeAuthor_Z721C83C5", "ScholarlyArticleConversion_splitAuthors_Z721C83C5", "ScholarlyArticleConversion_composeAuthors_Z721C83C5", "ScholarlyArticleConversion_decomposeAuthor_Z2F770004", "ScholarlyArticleConversion_decomposeAuthors_1AAAE9A5", "ScholarlyArticleConversion_composeScholarlyArticle_D324A6D", "ScholarlyArticleConversion_decomposeScholarlyArticle_Z6839B9E8"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ...arctrl_core.comment import Comment
from ...arctrl_core.Helper.collections_ import Option_fromValueWithDefault
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.Process.component import Component
from ...arctrl_core.Process.protocol import Protocol
from ...arctrl_core.Process.protocol_parameter import ProtocolParameter
from ...fable_library.list import (choose, singleton, of_array, FSharpList, empty, append)
from ...fable_library.option import (map, default_arg)
from ...fable_library.seq import map as map_1
from ...fable_library.string_ import replace
from ...fable_library.util import IEnumerable_1
from ...thoth_json_core.decode import (object, list_1 as list_1_2, IOptionalGetter, string, IGetters)
from ...thoth_json_core.encode import list_1 as list_1_1
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from ..comment import (ROCrate_encoder as ROCrate_encoder_2, ROCrate_decoder as ROCrate_decoder_2, ISAJson_encoder as ISAJson_encoder_3, ISAJson_decoder as ISAJson_decoder_3)
from ..context.rocrate.isa_protocol_context import context_jsonvalue
from ..decode import Decode_uri
from ..encode import (try_include, try_include_list_opt)
from ..idtable import encode
from ..ontology_annotation import (OntologyAnnotation_ROCrate_encoderDefinedTerm, OntologyAnnotation_ROCrate_decoderDefinedTerm, OntologyAnnotation_ISAJson_encoder, OntologyAnnotation_ISAJson_decoder)
from .component import (ROCrate_encoder as ROCrate_encoder_1, ROCrate_decoder as ROCrate_decoder_1, ISAJson_encoder as ISAJson_encoder_2, ISAJson_decoder as ISAJson_decoder_2)
from .protocol_parameter import (ISAJson_encoder as ISAJson_encoder_1, ISAJson_decoder as ISAJson_decoder_1)

__A_ = TypeVar("__A_")

def ROCrate_genID(study_name: str | None, assay_name: str | None, process_name: str | None, p: Protocol) -> str:
    match_value: str | None = p.ID
    (pattern_matching_result, id_1) = (None, None)
    if match_value is not None:
        if match_value != "":
            pattern_matching_result = 0
            id_1 = match_value

        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        return id_1

    elif pattern_matching_result == 1:
        match_value_1: str | None = p.Uri
        if match_value_1 is None:
            match_value_2: str | None = p.Name
            if match_value_2 is None:
                (pattern_matching_result_1, an, pn, sn, pn_1, sn_1, pn_2) = (None, None, None, None, None, None, None)
                if study_name is None:
                    if assay_name is None:
                        if process_name is not None:
                            pattern_matching_result_1 = 2
                            pn_2 = process_name

                        else: 
                            pattern_matching_result_1 = 3


                    else: 
                        pattern_matching_result_1 = 3


                elif assay_name is None:
                    if process_name is not None:
                        pattern_matching_result_1 = 1
                        pn_1 = process_name
                        sn_1 = study_name

                    else: 
                        pattern_matching_result_1 = 3


                elif process_name is not None:
                    pattern_matching_result_1 = 0
                    an = assay_name
                    pn = process_name
                    sn = study_name

                else: 
                    pattern_matching_result_1 = 3

                if pattern_matching_result_1 == 0:
                    return (((("#Protocol_" + replace(sn, " ", "_")) + "_") + replace(an, " ", "_")) + "_") + replace(pn, " ", "_")

                elif pattern_matching_result_1 == 1:
                    return (("#Protocol_" + replace(sn_1, " ", "_")) + "_") + replace(pn_1, " ", "_")

                elif pattern_matching_result_1 == 2:
                    return "#Protocol_" + replace(pn_2, " ", "_")

                elif pattern_matching_result_1 == 3:
                    return "#EmptyProtocol"


            else: 
                return "#Protocol_" + replace(match_value_2, " ", "_")


        else: 
            return match_value_1




def ROCrate_encoder(study_name: str | None, assay_name: str | None, process_name: str | None, oa: Protocol) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], study_name: Any=study_name, assay_name: Any=assay_name, process_name: Any=process_name, oa: Any=oa) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2496(__unit: None=None, study_name: Any=study_name, assay_name: Any=assay_name, process_name: Any=process_name, oa: Any=oa) -> IEncodable:
        value: str = ROCrate_genID(study_name, assay_name, process_name, oa)
        class ObjectExpr2495(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2495()

    class ObjectExpr2497(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], study_name: Any=study_name, assay_name: Any=assay_name, process_name: Any=process_name, oa: Any=oa) -> Any:
            return helpers_1.encode_string("Protocol")

    def _arrow2499(value_2: str, study_name: Any=study_name, assay_name: Any=assay_name, process_name: Any=process_name, oa: Any=oa) -> IEncodable:
        class ObjectExpr2498(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_2)

        return ObjectExpr2498()

    def _arrow2500(oa_1: OntologyAnnotation, study_name: Any=study_name, assay_name: Any=assay_name, process_name: Any=process_name, oa: Any=oa) -> IEncodable:
        return OntologyAnnotation_ROCrate_encoderDefinedTerm(oa_1)

    def _arrow2502(value_4: str, study_name: Any=study_name, assay_name: Any=assay_name, process_name: Any=process_name, oa: Any=oa) -> IEncodable:
        class ObjectExpr2501(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_4)

        return ObjectExpr2501()

    def _arrow2504(value_6: str, study_name: Any=study_name, assay_name: Any=assay_name, process_name: Any=process_name, oa: Any=oa) -> IEncodable:
        class ObjectExpr2503(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
                return helpers_4.encode_string(value_6)

        return ObjectExpr2503()

    def _arrow2506(value_8: str, study_name: Any=study_name, assay_name: Any=assay_name, process_name: Any=process_name, oa: Any=oa) -> IEncodable:
        class ObjectExpr2505(IEncodable):
            def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
                return helpers_5.encode_string(value_8)

        return ObjectExpr2505()

    def _arrow2507(comment: Comment, study_name: Any=study_name, assay_name: Any=assay_name, process_name: Any=process_name, oa: Any=oa) -> IEncodable:
        return ROCrate_encoder_2(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("@id", _arrow2496()), ("@type", list_1_1(singleton(ObjectExpr2497()))), try_include("name", _arrow2499, oa.Name), try_include("protocolType", _arrow2500, oa.ProtocolType), try_include("description", _arrow2502, oa.Description), try_include("uri", _arrow2504, oa.Uri), try_include("version", _arrow2506, oa.Version), try_include_list_opt("components", ROCrate_encoder_1, oa.Components), try_include_list_opt("comments", _arrow2507, oa.Comments), ("@context", context_jsonvalue)]))
    class ObjectExpr2508(IEncodable):
        def Encode(self, helpers_6: IEncoderHelpers_1[Any], study_name: Any=study_name, assay_name: Any=assay_name, process_name: Any=process_name, oa: Any=oa) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_6))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_6.encode_object(arg)

    return ObjectExpr2508()


def _arrow2520(get: IGetters) -> Protocol:
    def _arrow2512(__unit: None=None) -> FSharpList[Component]:
        list_4: FSharpList[Component]
        def _arrow2509(__unit: None=None) -> FSharpList[Component] | None:
            arg_1: Decoder_1[FSharpList[Component]] = list_1_2(ROCrate_decoder_1)
            object_arg: IOptionalGetter = get.Optional
            return object_arg.Field("components", arg_1)

        list_2: FSharpList[Component] = default_arg(_arrow2509(), empty())
        def _arrow2510(__unit: None=None) -> FSharpList[Component] | None:
            arg_3: Decoder_1[FSharpList[Component]] = list_1_2(ROCrate_decoder_1)
            object_arg_1: IOptionalGetter = get.Optional
            return object_arg_1.Field("reagents", arg_3)

        list_4 = append(default_arg(_arrow2510(), empty()), list_2)
        def _arrow2511(__unit: None=None) -> FSharpList[Component] | None:
            arg_5: Decoder_1[FSharpList[Component]] = list_1_2(ROCrate_decoder_1)
            object_arg_2: IOptionalGetter = get.Optional
            return object_arg_2.Field("computationalTools", arg_5)

        return append(default_arg(_arrow2511(), empty()), list_4)

    components: FSharpList[Component] | None = Option_fromValueWithDefault(empty(), _arrow2512())
    def _arrow2513(__unit: None=None) -> str | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("@id", Decode_uri)

    def _arrow2514(__unit: None=None) -> str | None:
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("name", string)

    def _arrow2515(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("protocolType", OntologyAnnotation_ROCrate_decoderDefinedTerm)

    def _arrow2516(__unit: None=None) -> str | None:
        object_arg_6: IOptionalGetter = get.Optional
        return object_arg_6.Field("description", string)

    def _arrow2517(__unit: None=None) -> str | None:
        object_arg_7: IOptionalGetter = get.Optional
        return object_arg_7.Field("uri", Decode_uri)

    def _arrow2518(__unit: None=None) -> str | None:
        object_arg_8: IOptionalGetter = get.Optional
        return object_arg_8.Field("version", string)

    def _arrow2519(__unit: None=None) -> FSharpList[Comment] | None:
        arg_19: Decoder_1[FSharpList[Comment]] = list_1_2(ROCrate_decoder_2)
        object_arg_9: IOptionalGetter = get.Optional
        return object_arg_9.Field("comments", arg_19)

    return Protocol(_arrow2513(), _arrow2514(), _arrow2515(), _arrow2516(), _arrow2517(), _arrow2518(), None, components, _arrow2519())


ROCrate_decoder: Decoder_1[Protocol] = object(_arrow2520)

def ISAJson_encoder(study_name: str | None, assay_name: str | None, process_name: str | None, id_map: Any | None, oa: Protocol) -> IEncodable:
    def f(oa_1: Protocol, study_name: Any=study_name, assay_name: Any=assay_name, process_name: Any=process_name, id_map: Any=id_map, oa: Any=oa) -> IEncodable:
        def chooser(tupled_arg: tuple[str, IEncodable | None], oa_1: Any=oa_1) -> tuple[str, IEncodable] | None:
            def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
                return (tupled_arg[0], v_1)

            return map(mapping, tupled_arg[1])

        def _arrow2524(value: str, oa_1: Any=oa_1) -> IEncodable:
            class ObjectExpr2523(IEncodable):
                def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                    return helpers.encode_string(value)

            return ObjectExpr2523()

        def _arrow2526(value_2: str, oa_1: Any=oa_1) -> IEncodable:
            class ObjectExpr2525(IEncodable):
                def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_1.encode_string(value_2)

            return ObjectExpr2525()

        def _arrow2527(oa_2: OntologyAnnotation, oa_1: Any=oa_1) -> IEncodable:
            return OntologyAnnotation_ISAJson_encoder(id_map, oa_2)

        def _arrow2529(value_4: str, oa_1: Any=oa_1) -> IEncodable:
            class ObjectExpr2528(IEncodable):
                def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_2.encode_string(value_4)

            return ObjectExpr2528()

        def _arrow2531(value_6: str, oa_1: Any=oa_1) -> IEncodable:
            class ObjectExpr2530(IEncodable):
                def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_3.encode_string(value_6)

            return ObjectExpr2530()

        def _arrow2533(value_8: str, oa_1: Any=oa_1) -> IEncodable:
            class ObjectExpr2532(IEncodable):
                def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_4.encode_string(value_8)

            return ObjectExpr2532()

        def _arrow2534(value_10: ProtocolParameter, oa_1: Any=oa_1) -> IEncodable:
            return ISAJson_encoder_1(id_map, value_10)

        def _arrow2535(c: Component, oa_1: Any=oa_1) -> IEncodable:
            return ISAJson_encoder_2(id_map, c)

        def _arrow2536(comment: Comment, oa_1: Any=oa_1) -> IEncodable:
            return ISAJson_encoder_3(id_map, comment)

        values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("@id", _arrow2524, ROCrate_genID(study_name, assay_name, process_name, oa_1)), try_include("name", _arrow2526, oa_1.Name), try_include("protocolType", _arrow2527, oa_1.ProtocolType), try_include("description", _arrow2529, oa_1.Description), try_include("uri", _arrow2531, oa_1.Uri), try_include("version", _arrow2533, oa_1.Version), try_include_list_opt("parameters", _arrow2534, oa_1.Parameters), try_include_list_opt("components", _arrow2535, oa_1.Components), try_include_list_opt("comments", _arrow2536, oa_1.Comments)]))
        class ObjectExpr2537(IEncodable):
            def Encode(self, helpers_5: IEncoderHelpers_1[Any], oa_1: Any=oa_1) -> Any:
                def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                    return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_5))

                arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
                return helpers_5.encode_object(arg)

        return ObjectExpr2537()

    if id_map is not None:
        def _arrow2538(p_1: Protocol, study_name: Any=study_name, assay_name: Any=assay_name, process_name: Any=process_name, id_map: Any=id_map, oa: Any=oa) -> str:
            return ROCrate_genID(study_name, assay_name, process_name, p_1)

        return encode(_arrow2538, f, oa, id_map)

    else: 
        return f(oa)



def _arrow2548(get: IGetters) -> Protocol:
    def _arrow2539(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("@id", Decode_uri)

    def _arrow2540(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("name", string)

    def _arrow2541(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("protocolType", OntologyAnnotation_ISAJson_decoder)

    def _arrow2542(__unit: None=None) -> str | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("description", string)

    def _arrow2543(__unit: None=None) -> str | None:
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("uri", Decode_uri)

    def _arrow2544(__unit: None=None) -> str | None:
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("version", string)

    def _arrow2545(__unit: None=None) -> FSharpList[ProtocolParameter] | None:
        arg_13: Decoder_1[FSharpList[ProtocolParameter]] = list_1_2(ISAJson_decoder_1)
        object_arg_6: IOptionalGetter = get.Optional
        return object_arg_6.Field("parameters", arg_13)

    def _arrow2546(__unit: None=None) -> FSharpList[Component] | None:
        arg_15: Decoder_1[FSharpList[Component]] = list_1_2(ISAJson_decoder_2)
        object_arg_7: IOptionalGetter = get.Optional
        return object_arg_7.Field("components", arg_15)

    def _arrow2547(__unit: None=None) -> FSharpList[Comment] | None:
        arg_17: Decoder_1[FSharpList[Comment]] = list_1_2(ISAJson_decoder_3)
        object_arg_8: IOptionalGetter = get.Optional
        return object_arg_8.Field("comments", arg_17)

    return Protocol(_arrow2539(), _arrow2540(), _arrow2541(), _arrow2542(), _arrow2543(), _arrow2544(), _arrow2545(), _arrow2546(), _arrow2547())


ISAJson_decoder: Decoder_1[Protocol] = object(_arrow2548)

__all__ = ["ROCrate_genID", "ROCrate_encoder", "ROCrate_decoder", "ISAJson_encoder", "ISAJson_decoder"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ...arctrl_core.Helper.collections_ import Option_fromValueWithDefault
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.Process.protocol_parameter import ProtocolParameter
from ...fable_library.list import (choose, of_array, FSharpList)
from ...fable_library.option import (default_arg, map as map_1)
from ...fable_library.seq import map as map_2
from ...fable_library.util import IEnumerable_1
from ...thoth_json_core.decode import (map, object, IOptionalGetter, IGetters)
from ...thoth_json_core.types import (IEncodable, Decoder_1, IEncoderHelpers_1)
from ..encode import try_include
from ..idtable import encode
from ..ontology_annotation import (OntologyAnnotation_encoder, OntologyAnnotation_decoder, OntologyAnnotation_ROCrate_genID, OntologyAnnotation_ISAJson_encoder, OntologyAnnotation_ISAJson_decoder)

__A_ = TypeVar("__A_")

def encoder(value: ProtocolParameter) -> IEncodable:
    return OntologyAnnotation_encoder(default_arg(value.ParameterName, OntologyAnnotation()))


def ctor(oa: OntologyAnnotation) -> ProtocolParameter:
    oa_1: OntologyAnnotation | None = Option_fromValueWithDefault(OntologyAnnotation(), oa)
    return ProtocolParameter.create(None, oa_1)


decoder: Decoder_1[ProtocolParameter] = map(ctor, OntologyAnnotation_decoder)

def ISAJson_genID(p: ProtocolParameter) -> str:
    match_value: OntologyAnnotation | None = p.ParameterName
    if match_value is None:
        return "#EmptyProtocolParameter"

    else: 
        return ("#ProtocolParameter/" + OntologyAnnotation_ROCrate_genID(match_value)) + ""



def ISAJson_encoder(id_map: Any | None, value: ProtocolParameter) -> IEncodable:
    def f(value_1: ProtocolParameter, id_map: Any=id_map, value: Any=value) -> IEncodable:
        def chooser(tupled_arg: tuple[str, IEncodable | None], value_1: Any=value_1) -> tuple[str, IEncodable] | None:
            def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
                return (tupled_arg[0], v_1)

            return map_1(mapping, tupled_arg[1])

        def _arrow2445(value_2: str, value_1: Any=value_1) -> IEncodable:
            class ObjectExpr2444(IEncodable):
                def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                    return helpers.encode_string(value_2)

            return ObjectExpr2444()

        def _arrow2446(oa: OntologyAnnotation, value_1: Any=value_1) -> IEncodable:
            return OntologyAnnotation_ISAJson_encoder(id_map, oa)

        values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("@id", _arrow2445, ISAJson_genID(value_1)), try_include("parameterName", _arrow2446, value_1.ParameterName)]))
        class ObjectExpr2447(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any], value_1: Any=value_1) -> Any:
                def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                    return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_1))

                arg: IEnumerable_1[tuple[str, __A_]] = map_2(mapping_1, values)
                return helpers_1.encode_object(arg)

        return ObjectExpr2447()

    if id_map is not None:
        def _arrow2448(p_1: ProtocolParameter, id_map: Any=id_map, value: Any=value) -> str:
            return ISAJson_genID(p_1)

        return encode(_arrow2448, f, value, id_map)

    else: 
        return f(value)



def _arrow2450(get: IGetters) -> ProtocolParameter:
    def _arrow2449(__unit: None=None) -> OntologyAnnotation | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("parameterName", OntologyAnnotation_ISAJson_decoder)

    return ProtocolParameter(None, _arrow2449())


ISAJson_decoder: Decoder_1[ProtocolParameter] = object(_arrow2450)

__all__ = ["encoder", "decoder", "ISAJson_genID", "ISAJson_encoder", "ISAJson_decoder"]




from __future__ import annotations
from abc import abstractmethod
from collections.abc import Callable
from typing import (Protocol, Generic, TypeVar)

_T = TypeVar("_T")

class Awaitable_1(Protocol, Generic[_T]):
    pass

class Future_1(Awaitable_1, Generic[_T]):
    @abstractmethod
    def add_done_callback(self, callback: Callable[[], None]) -> None:
        ...

    @abstractmethod
    def cancel(self) -> None:
        ...

    @abstractmethod
    def cancelled(self) -> bool:
        ...

    @abstractmethod
    def done(self) -> bool:
        ...

    @abstractmethod
    def exception(self) -> Exception:
        ...

    @abstractmethod
    def get_name(self) -> str:
        ...

    @abstractmethod
    def result(self) -> _T:
        ...

    @abstractmethod
    def set_exception(self, __exception: Exception) -> None:
        ...

    @abstractmethod
    def set_name(self, __arg0: str) -> None:
        ...

    @abstractmethod
    def set_result(self, __result: _T) -> None:
        ...





from __future__ import annotations
from typing import Any
from ...fable_library.list import (FSharpList, choose, empty)
from ...fable_library.option import (default_arg, map)
from ...fable_library.reflection import (TypeInfo, union_type)
from ...fable_library.string_ import (to_text, printf)
from ...fable_library.types import (to_string, Array, Union)
from ..data import (Data_reflection, Data)
from ..data_file import DataFile
from ..ontology_annotation import OntologyAnnotation
from .factor_value import FactorValue
from .material import (Material_reflection, Material, Material_getUnits_ZBCDEB61, Material_create_Z66909A6D)
from .material_attribute import MaterialAttribute
from .material_attribute_value import MaterialAttributeValue
from .sample import (Sample_reflection, Sample_get_empty, Sample, Sample_setFactorValues, Sample_getUnits_53716792, Sample_create_62424AD2)
from .source import Source

def _expr768() -> TypeInfo:
    return union_type("ARCtrl.Process.ProcessOutput", [], ProcessOutput, lambda: [[("Item", Sample_reflection())], [("Item", Data_reflection())], [("Item", Material_reflection())]])


class ProcessOutput(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Sample", "Data", "Material"]

    def Print(self, __unit: None=None) -> str:
        this: ProcessOutput = self
        return to_string(this)

    def PrintCompact(self, __unit: None=None) -> str:
        this: ProcessOutput = self
        if this.tag == 2:
            arg_1: str = this.fields[0].PrintCompact()
            return to_text(printf("Material {%s}"))(arg_1)

        elif this.tag == 1:
            arg_2: str = this.fields[0].PrintCompact()
            return to_text(printf("Data {%s}"))(arg_2)

        else: 
            arg: str = this.fields[0].PrintCompact()
            return to_text(printf("Sample {%s}"))(arg)



ProcessOutput_reflection = _expr768

def ProcessOutput__get_TryName(this: ProcessOutput) -> str | None:
    if this.tag == 2:
        return this.fields[0].Name

    elif this.tag == 1:
        return this.fields[0].Name

    else: 
        return this.fields[0].Name



def ProcessOutput__get_Name(this: ProcessOutput) -> str:
    return default_arg(ProcessOutput__get_TryName(this), "")


def ProcessOutput_get_Default(__unit: None=None) -> ProcessOutput:
    return ProcessOutput(0, Sample_get_empty())


def ProcessOutput_tryGetName_Z42C11600(po: ProcessOutput) -> str | None:
    return ProcessOutput__get_TryName(po)


def ProcessOutput_getName_Z42C11600(po: ProcessOutput) -> str:
    return ProcessOutput__get_Name(po)


def ProcessOutput_nameEquals(name: str, po: ProcessOutput) -> bool:
    return ProcessOutput__get_Name(po) == name


def ProcessOutput_isSample_Z42C11600(po: ProcessOutput) -> bool:
    if po.tag == 0:
        return True

    else: 
        return False



def ProcessOutput_isData_Z42C11600(po: ProcessOutput) -> bool:
    if po.tag == 1:
        return True

    else: 
        return False



def ProcessOutput_isMaterial_Z42C11600(po: ProcessOutput) -> bool:
    if po.tag == 2:
        return True

    else: 
        return False



def ProcessOutput__isSample(this: ProcessOutput) -> bool:
    return ProcessOutput_isSample_Z42C11600(this)


def ProcessOutput__isData(this: ProcessOutput) -> bool:
    return ProcessOutput_isData_Z42C11600(this)


def ProcessOutput__isMaterial(this: ProcessOutput) -> bool:
    return ProcessOutput_isMaterial_Z42C11600(this)


def ProcessOutput_trySample_Z42C11600(po: ProcessOutput) -> Sample | None:
    if po.tag == 0:
        return po.fields[0]

    else: 
        return None



def ProcessOutput_tryData_Z42C11600(po: ProcessOutput) -> Data | None:
    if po.tag == 1:
        return po.fields[0]

    else: 
        return None



def ProcessOutput_tryMaterial_Z42C11600(po: ProcessOutput) -> Material | None:
    if po.tag == 2:
        return po.fields[0]

    else: 
        return None



def ProcessOutput_tryGetCharacteristicValues_Z42C11600(po: ProcessOutput) -> FSharpList[MaterialAttributeValue] | None:
    if po.tag == 2:
        return po.fields[0].Characteristics

    elif po.tag == 1:
        return None

    else: 
        return po.fields[0].Characteristics



def ProcessOutput_tryGetCharacteristics_Z42C11600(po: ProcessOutput) -> FSharpList[MaterialAttribute] | None:
    def mapping(list_1: FSharpList[MaterialAttributeValue], po: Any=po) -> FSharpList[MaterialAttribute]:
        def chooser(c: MaterialAttributeValue, list_1: Any=list_1) -> MaterialAttribute | None:
            return c.Category

        return choose(chooser, list_1)

    return map(mapping, ProcessOutput_tryGetCharacteristicValues_Z42C11600(po))


def ProcessOutput_tryGetFactorValues_Z42C11600(po: ProcessOutput) -> FSharpList[FactorValue] | None:
    if po.tag == 2:
        return None

    elif po.tag == 1:
        return None

    else: 
        return po.fields[0].FactorValues



def ProcessOutput_setFactorValues(values: FSharpList[FactorValue], po: ProcessOutput) -> ProcessOutput:
    if po.tag == 2:
        return po

    elif po.tag == 1:
        return po

    else: 
        return ProcessOutput(0, Sample_setFactorValues(values, po.fields[0]))



def ProcessOutput_getFactorValues_Z42C11600(po: ProcessOutput) -> FSharpList[FactorValue]:
    return default_arg(ProcessOutput_tryGetFactorValues_Z42C11600(po), empty())


def ProcessOutput_getUnits_Z42C11600(po: ProcessOutput) -> FSharpList[OntologyAnnotation]:
    if po.tag == 2:
        return Material_getUnits_ZBCDEB61(po.fields[0])

    elif po.tag == 1:
        return empty()

    else: 
        return Sample_getUnits_53716792(po.fields[0])



def ProcessOutput_createSample_Z598187B7(name: str, characteristics: FSharpList[MaterialAttributeValue] | None=None, factors: FSharpList[FactorValue] | None=None, derives_from: FSharpList[Source] | None=None) -> ProcessOutput:
    return ProcessOutput(0, Sample_create_62424AD2(None, name, characteristics, factors, derives_from))


def ProcessOutput_createMaterial_4452CB4C(name: str, characteristics: FSharpList[MaterialAttributeValue] | None=None, derives_from: FSharpList[Material] | None=None) -> ProcessOutput:
    return ProcessOutput(2, Material_create_Z66909A6D(None, name, None, characteristics, derives_from))


def ProcessOutput_createImageFile_Z721C83C5(name: str) -> ProcessOutput:
    return ProcessOutput(1, Data.create(None, name, DataFile(2)))


def ProcessOutput_createRawData_Z721C83C5(name: str) -> ProcessOutput:
    return ProcessOutput(1, Data.create(None, name, DataFile(0)))


def ProcessOutput_createDerivedData_Z721C83C5(name: str) -> ProcessOutput:
    return ProcessOutput(1, Data.create(None, name, DataFile(1)))


__all__ = ["ProcessOutput_reflection", "ProcessOutput__get_TryName", "ProcessOutput__get_Name", "ProcessOutput_get_Default", "ProcessOutput_tryGetName_Z42C11600", "ProcessOutput_getName_Z42C11600", "ProcessOutput_nameEquals", "ProcessOutput_isSample_Z42C11600", "ProcessOutput_isData_Z42C11600", "ProcessOutput_isMaterial_Z42C11600", "ProcessOutput__isSample", "ProcessOutput__isData", "ProcessOutput__isMaterial", "ProcessOutput_trySample_Z42C11600", "ProcessOutput_tryData_Z42C11600", "ProcessOutput_tryMaterial_Z42C11600", "ProcessOutput_tryGetCharacteristicValues_Z42C11600", "ProcessOutput_tryGetCharacteristics_Z42C11600", "ProcessOutput_tryGetFactorValues_Z42C11600", "ProcessOutput_setFactorValues", "ProcessOutput_getFactorValues_Z42C11600", "ProcessOutput_getUnits_Z42C11600", "ProcessOutput_createSample_Z598187B7", "ProcessOutput_createMaterial_4452CB4C", "ProcessOutput_createImageFile_Z721C83C5", "ProcessOutput_createRawData_Z721C83C5", "ProcessOutput_createDerivedData_Z721C83C5"]




from __future__ import annotations
from abc import abstractmethod
from typing import Protocol
from ...fable_library.list import FSharpList
from ...fable_library.types import (Array, int64)

class ITimeSeries(Protocol):
    @abstractmethod
    def plot(self, start: str, end: str, aggregates: Array[str], granularity: str) -> None:
        ...


class ITimeSeriesApi(Protocol):
    @abstractmethod
    def list(self) -> FSharpList[ITimeSeries]:
        ...

    @abstractmethod
    def retrieve(self, id: int64) -> ITimeSeries:
        ...





from __future__ import annotations
from typing import Any
from ...arctrl_core.comment import Comment
from ...arctrl_core.Helper.regex import ActivePatterns__007CRegex_007C__007C
from ...fable_library.option import (default_arg, bind)
from ...fable_library.reg_exp import (get_item, groups)
from ...fable_library.string_ import (to_text, printf)
from .conversions import Option_fromValueWithDefault

Comment_commentValueKey: str = "commentValue"

Comment_commentPattern: str = ("Comment\\s*\\[<(?<" + Comment_commentValueKey) + ">.+)>\\]"

Comment_commentPatternNoAngleBrackets: str = ("Comment\\s*\\[(?<" + Comment_commentValueKey) + ">.+)\\]"

def Comment__007CComment_007C__007C(key: str) -> str | None:
    active_pattern_result: Any | None = ActivePatterns__007CRegex_007C__007C(Comment_commentPattern, key)
    if active_pattern_result is not None:
        r: Any = active_pattern_result
        return get_item(groups(r), Comment_commentValueKey) or ""

    else: 
        active_pattern_result_1: Any | None = ActivePatterns__007CRegex_007C__007C(Comment_commentPatternNoAngleBrackets, key)
        if active_pattern_result_1 is not None:
            r_1: Any = active_pattern_result_1
            v: str = get_item(groups(r_1), Comment_commentValueKey) or ""
            if v == "<>":
                return None

            else: 
                return v


        else: 
            return None




def Comment_wrapCommentKey(k: str) -> str:
    return to_text(printf("Comment[%s]"))(k)


def Comment_fromString(k: str, v: str) -> Comment:
    name: str | None = Option_fromValueWithDefault("", k)
    value: str | None = Option_fromValueWithDefault("", v)
    return Comment.make(name, value)


def Comment_toString(c: Comment) -> tuple[str, str]:
    return (default_arg(c.Name, ""), default_arg(c.Value, ""))


Remark_remarkValueKey: str = "remarkValue"

Remark_remarkPattern: str = ("#(?<" + Remark_remarkValueKey) + ">.+)"

def Remark__007CRemark_007C__007C(key: str | None=None) -> str | None:
    def binder(k: str, key: Any=key) -> str | None:
        active_pattern_result: Any | None = ActivePatterns__007CRegex_007C__007C(Remark_remarkPattern, k)
        if active_pattern_result is not None:
            r: Any = active_pattern_result
            return get_item(groups(r), Remark_remarkValueKey) or ""

        else: 
            return None


    return bind(binder, key)


def Remark_wrapRemark(r: str) -> str:
    return to_text(printf("#%s"))(r)


__all__ = ["Comment_commentValueKey", "Comment_commentPattern", "Comment_commentPatternNoAngleBrackets", "Comment__007CComment_007C__007C", "Comment_wrapCommentKey", "Comment_fromString", "Comment_toString", "Remark_remarkValueKey", "Remark_remarkPattern", "Remark__007CRemark_007C__007C", "Remark_wrapRemark"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ..fable_library.array_ import (initialize, skip as skip_1, try_item)
from ..fable_library.list import (try_pick, FSharpList, is_empty, reverse, head, cons, tail, empty)
from ..fable_library.map_util import try_get_value
from ..fable_library.option import (value, some)
from ..fable_library.seq import (skip, max_by, try_find)
from ..fable_library.types import (Array, FSharpRef)
from ..fable_library.util import (IEnumerable_1, compare_primitives)

__A = TypeVar("__A")

_T = TypeVar("_T")

_A = TypeVar("_A")

_B = TypeVar("_B")

_C = TypeVar("_C")

_D = TypeVar("_D")

_U = TypeVar("_U")

__B = TypeVar("__B")

__C = TypeVar("__C")

__D = TypeVar("__D")

_V = TypeVar("_V")

def Seq_trySkip(i: int, s: IEnumerable_1[Any]) -> IEnumerable_1[Any] | None:
    try: 
        return skip(i, s)

    except Exception as match_value:
        return None



def Array_ofIndexedSeq(s: IEnumerable_1[tuple[int, str]]) -> Array[str]:
    def _arrow3265(tuple: tuple[int, str], s: Any=s) -> int:
        return tuple[0]

    class ObjectExpr3266:
        @property
        def Compare(self) -> Callable[[int, int], int]:
            return compare_primitives

    def _arrow3268(i: int, s: Any=s) -> str:
        def _arrow3267(arg: tuple[int, str]) -> bool:
            return i == arg[0]

        match_value: tuple[int, str] | None = try_find(_arrow3267, s)
        if match_value is None:
            return ""

        else: 
            i_1: int = match_value[0] or 0
            return match_value[1]


    return initialize(1 + max_by(_arrow3265, s, ObjectExpr3266())[0], _arrow3268, None)


def Array_trySkip(i: int, a: Array[Any]) -> Array[Any] | None:
    try: 
        return skip_1(i, a, None)

    except Exception as match_value:
        return None



def Array_tryItemDefault(i: int, d: Any, a: Array[Any]) -> Any:
    match_value: __A | None = try_item(i, a)
    if match_value is None:
        return d

    else: 
        return value(match_value)



def Array_map4(f: Callable[[_A, _B, _C, _D], _T], aa: Array[Any], ba: Array[Any], ca: Array[Any], da: Array[Any]) -> Array[Any]:
    if not ((len(ca) == len(da)) if ((len(ba) == len(ca)) if (len(aa) == len(ba)) else False) else False):
        raise Exception("")

    def _arrow3269(i: int, f: Any=f, aa: Any=aa, ba: Any=ba, ca: Any=ca, da: Any=da) -> _T:
        return f(aa[i], ba[i], ca[i], da[i])

    return initialize(len(aa), _arrow3269, None)


def List_tryPickDefault(chooser: Callable[[_T], _U | None], d: Any, list_1: FSharpList[Any]) -> Any:
    match_value: _U | None = try_pick(chooser, list_1)
    if match_value is None:
        return d

    else: 
        return value(match_value)



def List_unzip4(l: FSharpList[tuple[_A, _B, _C, _D]]) -> tuple[FSharpList[_A], FSharpList[_B], FSharpList[_C], FSharpList[_D]]:
    def loop(la_mut: FSharpList[Any], lb_mut: FSharpList[Any], lc_mut: FSharpList[Any], ld_mut: FSharpList[Any], l_1_mut: FSharpList[tuple[__A, __B, __C, __D]], l: Any=l) -> tuple[FSharpList[__A], FSharpList[__B], FSharpList[__C], FSharpList[__D]]:
        while True:
            (la, lb, lc, ld, l_1) = (la_mut, lb_mut, lc_mut, ld_mut, l_1_mut)
            if is_empty(l_1):
                return (reverse(la), reverse(lb), reverse(lc), reverse(ld))

            else: 
                la_mut = cons(head(l_1)[0], la)
                lb_mut = cons(head(l_1)[1], lb)
                lc_mut = cons(head(l_1)[2], lc)
                ld_mut = cons(head(l_1)[3], ld)
                l_1_mut = tail(l_1)
                continue

            break

    return loop(empty(), empty(), empty(), empty(), l)


def Dictionary_tryGetValue(k: Any, dict_1: Any) -> Any | None:
    pattern_input: tuple[bool, _V]
    out_arg: _V = None
    def _arrow3270(__unit: None=None, k: Any=k, dict_1: Any=dict_1) -> _V:
        return out_arg

    def _arrow3271(v: _V | None=None, k: Any=k, dict_1: Any=dict_1) -> None:
        nonlocal out_arg
        out_arg = v

    pattern_input = (try_get_value(dict_1, k, FSharpRef(_arrow3270, _arrow3271)), out_arg)
    if pattern_input[0]:
        return some(pattern_input[1])

    else: 
        return None



def Dictionary_tryGetString(k: Any, dict_1: Any) -> str | None:
    pattern_input: tuple[bool, str]
    out_arg: str = None
    def _arrow3272(__unit: None=None, k: Any=k, dict_1: Any=dict_1) -> str:
        return out_arg

    def _arrow3273(v: str, k: Any=k, dict_1: Any=dict_1) -> None:
        nonlocal out_arg
        out_arg = v

    pattern_input = (try_get_value(dict_1, k, FSharpRef(_arrow3272, _arrow3273)), out_arg)
    v_1: str = pattern_input[1]
    if (v_1.strip() != "") if pattern_input[0] else False:
        return v_1.strip()

    else: 
        return None



__all__ = ["Seq_trySkip", "Array_ofIndexedSeq", "Array_trySkip", "Array_tryItemDefault", "Array_map4", "List_tryPickDefault", "List_unzip4", "Dictionary_tryGetValue", "Dictionary_tryGetString"]




from __future__ import annotations


class UTF16LE:
    def get_bytes(self, string: str, index: int | None = None, count: int | None = None) -> bytes:
        return string.encode("utf-16le")

    def get_string(self, bytes: bytes, index: int | None = None, count: int | None = None) -> str:
        if index is None:
            return bytes.decode("utf-8")
        else:
            return bytes.decode("utf-8")[index:count]


class UTF8:
    def get_bytes(self, string: str, index: int | None = None, count: int | None = None) -> bytes:
        if index is None:
            return string.encode("utf-8")
        else:
            return string.encode("utf-8")[index:count]

    def get_string(self, bytes: bytes, index: int | None = None, count: int | None = None) -> str:
        if index is None:
            return bytes.decode("utf-8")
        else:
            return bytes.decode("utf-8")[index:count]


_UTF8 = UTF8()
_UTF16 = UTF16LE()


def get_unicode():
    return _UTF16


def get_utf8() -> UTF8:
    return _UTF8


__all__ = ["UTF8", "UTF16LE", "get_utf8", "get_unicode"]



from __future__ import annotations

import re
from collections.abc import Iterable
from enum import IntEnum
from re import Match
from typing import (
    TYPE_CHECKING,
    Any,
    NoReturn,
    TypeVar,
)

from .types import FSharpRef, Union
from .util import Array


_K = TypeVar("_K")
_V = TypeVar("_V")


class CaseRules(IntEnum):
    Ignore = 0
    LowerFirst = 1
    SnakeCase = 2
    SnakeCaseAllCaps = 3
    KebabCase = 4


def dashify(string: str, separator: str) -> str:
    def match(m: Match[str]) -> str:
        s = m.string
        if len(s) == 1:
            return s.lower()
        else:
            return s[0] + separator + s[1].lower()

    return re.sub(r"[a-z]?[A-Z]", match, string)


def change_case(string: str, case_rule: CaseRules) -> str:
    if case_rule == CaseRules.LowerFirst:
        return string[0].lower() + string[1:]

    if case_rule == CaseRules.SnakeCase:
        return dashify(string, "_")
    if case_rule == CaseRules.SnakeCaseAllCaps:
        return dashify(string, "_").upper()
    if case_rule == CaseRules.KebabCase:
        return dashify(string, "-")

    return string


if TYPE_CHECKING:

    class FSharpMap(dict[_K, _V]): ...

else:
    from .map import FSharpMap


def add_to_set(v: _V, st: set[_V]) -> bool:
    if v in st:
        return False

    st.add(v)
    return True


def add_to_dict(di: dict[_K, _V], k: _K, v: _V) -> None:
    if k in di:
        raise Exception("An item with the same key has already been added. Key: " + str(k))

    di[k] = v


def remove_from_dict(di: dict[_K, Any], k: _K) -> bool:
    if k in di:
        del di[k]
        return True

    return False


def try_get_value(map: FSharpMap[_K, _V], key: _K, default_value: FSharpRef[_V]) -> bool:
    if key in map.keys():
        default_value.contents = map[key]
        return True

    return False


def get_item_from_dict(map: dict[_K, _V], key: _K) -> _V:
    if key in map:
        return map[key]
    else:
        raise Exception(f"The given key '{key}' was not present in the dictionary.")


def contains_value(v: _V, map: dict[Any, _V]) -> bool:
    return v in map.values()


def key_value_list(fields: Iterable[Any], case_rule: CaseRules = CaseRules.Ignore):
    obj: dict[str, Any] = {}
    defined_case_rule = case_rule

    def fail(kvPair: Any) -> NoReturn:
        raise Exception("Cannot infer key and value of " + str(kvPair))

    def assign(key: str, case_rule: CaseRules, value: Any):
        key = change_case(key, case_rule)
        obj[key] = value

    for kv_pair in fields:
        case_rule = CaseRules.Ignore
        if kv_pair is None:
            fail(kv_pair)

        # Deflate unions and use the defined case rule
        if isinstance(kv_pair, Union):
            name = kv_pair.cases()[kv_pair.tag]
            kv_pair = name if len(kv_pair.fields) == 0 else [name, *kv_pair.fields]
            case_rule = defined_case_rule

        if isinstance(kv_pair, Array):
            length = len(kv_pair)
            if length == 0:
                fail(kv_pair)

            elif length == 1:
                assign(kv_pair[0], case_rule, True)
                break

            elif length == 2:
                value = kv_pair[1]
                assign(kv_pair[0], case_rule, value)
                break

            assign(kv_pair[0], case_rule, kv_pair[1:])

        elif isinstance(kv_pair, str):
            assign(kv_pair, case_rule, True)
        else:
            fail(kv_pair)

    return obj



from __future__ import annotations
from dataclasses import dataclass
from typing import (Any, TypeVar)
from ....fable_library.reflection import (TypeInfo, string_type, record_type)
from ....fable_library.seq import map
from ....fable_library.types import Record
from ....fable_library.util import (to_enumerable, IEnumerable_1)
from ....thoth_json_core.types import (IEncodable, IEncoderHelpers_1)

__A_ = TypeVar("__A_")

def _expr1663() -> TypeInfo:
    return record_type("ARCtrl.Json.ROCrateContext.Investigation.IContext", [], IContext, lambda: [("sdo", string_type), ("arc", string_type), ("Investigation", string_type), ("identifier", string_type), ("title", string_type), ("description", string_type), ("submission_date", string_type), ("public_release_date", string_type), ("publications", string_type), ("people", string_type), ("studies", string_type), ("ontology_source_references", string_type), ("comments", string_type), ("publications_003F", string_type), ("filename", string_type)])


@dataclass(eq = False, repr = False, slots = True)
class IContext(Record):
    sdo: str
    arc: str
    Investigation: str
    identifier: str
    title: str
    description: str
    submission_date: str
    public_release_date: str
    publications: str
    people: str
    studies: str
    ontology_source_references: str
    comments: str
    publications_003F: str
    filename: str

IContext_reflection = _expr1663

def _arrow1679(__unit: None=None) -> IEncodable:
    class ObjectExpr1664(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
            return helpers.encode_string("http://schema.org/")

    class ObjectExpr1665(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
            return helpers_1.encode_string("sdo:Dataset")

    class ObjectExpr1666(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
            return helpers_2.encode_string("sdo:identifier")

    class ObjectExpr1667(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
            return helpers_3.encode_string("sdo:name")

    class ObjectExpr1668(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
            return helpers_4.encode_string("sdo:additionalType")

    class ObjectExpr1669(IEncodable):
        def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
            return helpers_5.encode_string("sdo:description")

    class ObjectExpr1670(IEncodable):
        def Encode(self, helpers_6: IEncoderHelpers_1[Any]) -> Any:
            return helpers_6.encode_string("sdo:dateCreated")

    class ObjectExpr1671(IEncodable):
        def Encode(self, helpers_7: IEncoderHelpers_1[Any]) -> Any:
            return helpers_7.encode_string("sdo:datePublished")

    class ObjectExpr1672(IEncodable):
        def Encode(self, helpers_8: IEncoderHelpers_1[Any]) -> Any:
            return helpers_8.encode_string("sdo:citation")

    class ObjectExpr1673(IEncodable):
        def Encode(self, helpers_9: IEncoderHelpers_1[Any]) -> Any:
            return helpers_9.encode_string("sdo:creator")

    class ObjectExpr1674(IEncodable):
        def Encode(self, helpers_10: IEncoderHelpers_1[Any]) -> Any:
            return helpers_10.encode_string("sdo:hasPart")

    class ObjectExpr1675(IEncodable):
        def Encode(self, helpers_11: IEncoderHelpers_1[Any]) -> Any:
            return helpers_11.encode_string("sdo:mentions")

    class ObjectExpr1676(IEncodable):
        def Encode(self, helpers_12: IEncoderHelpers_1[Any]) -> Any:
            return helpers_12.encode_string("sdo:comment")

    class ObjectExpr1677(IEncodable):
        def Encode(self, helpers_13: IEncoderHelpers_1[Any]) -> Any:
            return helpers_13.encode_string("sdo:alternateName")

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("sdo", ObjectExpr1664()), ("Investigation", ObjectExpr1665()), ("identifier", ObjectExpr1666()), ("title", ObjectExpr1667()), ("additionalType", ObjectExpr1668()), ("description", ObjectExpr1669()), ("submissionDate", ObjectExpr1670()), ("publicReleaseDate", ObjectExpr1671()), ("publications", ObjectExpr1672()), ("people", ObjectExpr1673()), ("studies", ObjectExpr1674()), ("ontologySourceReferences", ObjectExpr1675()), ("comments", ObjectExpr1676()), ("filename", ObjectExpr1677())])
    class ObjectExpr1678(IEncodable):
        def Encode(self, helpers_14: IEncoderHelpers_1[Any]) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_14))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_14.encode_object(arg)

    return ObjectExpr1678()


context_jsonvalue: IEncodable = _arrow1679()

context_str: str = "\r\n{\r\n  \"@context\": {\r\n    \"sdo\": \"http://schema.org/\",\r\n    \"arc\": \"http://purl.org/nfdi4plants/ontology/\",\r\n\r\n    \"Investigation\": \"sdo:Dataset\",\r\n\r\n    \"identifier\" : \"sdo:identifier\",\r\n    \"title\": \"sdo:name\",\r\n    \"description\": \"sdo:description\",\r\n    \"submissionDate\": \"sdo:dateCreated\",\r\n    \"publicReleaseDate\": \"sdo:datePublished\",\r\n    \"publications\": \"sdo:citation\",\r\n    \"people\": \"sdo:creator\",\r\n    \"studies\": \"sdo:hasPart\",\r\n    \"ontologySourceReferences\": \"sdo:mentions\",\r\n    \"comments\": \"sdo:disambiguatingDescription\",\r\n\r\n    \"publications?\": \"sdo:SubjectOf?\",\r\n    \"filename\": \"sdo:alternateName\"\r\n  }\r\n}\r\n    "

__all__ = ["IContext_reflection", "context_jsonvalue", "context_str"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_core.comment import (Comment, Remark)
from ...arctrl_core.Helper.collections_ import (Option_fromValueWithDefault, ResizeArray_iter)
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.publication import Publication
from ...arctrl_core.uri import URIModule_fromString
from ...fable_library.list import (of_array, FSharpList, length, singleton, initialize, map as map_1, empty, iterate_indexed, cons, reverse)
from ...fable_library.map_util import add_to_dict
from ...fable_library.option import (map, default_arg)
from ...fable_library.seq2 import List_distinct
from ...fable_library.types import Array
from ...fable_library.util import (string_hash, IEnumerable_1, IEnumerator)
from .comment import (Comment_fromString, Comment_toString)
from .sparse_table import (SparseTable_GetEmptyComments_3ECCA699, SparseTable__TryGetValueDefault_5BAE6133, SparseTable__TryGetValue_11FD62A8, SparseTable, SparseTable_Create_Z2192E64B, SparseTable_FromRows_Z5579EC29, SparseTable_ToRows_759CAFC1)

pub_med_idlabel: str = "PubMed ID"

doi_label: str = "DOI"

author_list_label: str = "Author List"

title_label: str = "Title"

status_label: str = "Status"

status_term_accession_number_label: str = "Status Term Accession Number"

status_term_source_reflabel: str = "Status Term Source REF"

labels: FSharpList[str] = of_array([pub_med_idlabel, doi_label, author_list_label, title_label, status_label, status_term_accession_number_label, status_term_source_reflabel])

def from_string(pub_med_id: str | None, doi: str | None, author: str | None, title: str | None, status: str | None, status_term_source_ref: str | None, status_term_accession_number: str | None, comments: Array[Comment]) -> Publication:
    status_1: OntologyAnnotation = OntologyAnnotation(status, status_term_source_ref, status_term_accession_number)
    def _arrow3459(s: str, pub_med_id: Any=pub_med_id, doi: Any=doi, author: Any=author, title: Any=title, status: Any=status, status_term_source_ref: Any=status_term_source_ref, status_term_accession_number: Any=status_term_accession_number, comments: Any=comments) -> str:
        return URIModule_fromString(s)

    pub_med_id_1: str | None = map(_arrow3459, pub_med_id)
    status_2: OntologyAnnotation | None = Option_fromValueWithDefault(OntologyAnnotation(), status_1)
    return Publication.make(pub_med_id_1, doi, author, title, status_2, comments)


def from_sparse_table(matrix: SparseTable) -> FSharpList[Publication]:
    if (length(matrix.CommentKeys) != 0) if (matrix.ColumnCount == 0) else False:
        comments: Array[Comment] = SparseTable_GetEmptyComments_3ECCA699(matrix)
        def _arrow3460(__unit: None=None, matrix: Any=matrix) -> Publication:
            return_val: Publication = Publication.create()
            return_val.Comments = comments
            return return_val

        return singleton(_arrow3460())

    else: 
        def _arrow3461(i: int, matrix: Any=matrix) -> Publication:
            def mapping(k: str) -> Comment:
                return Comment_fromString(k, SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (k, i)))

            comments_1: Array[Comment] = list(map_1(mapping, matrix.CommentKeys))
            return from_string(SparseTable__TryGetValue_11FD62A8(matrix, (pub_med_idlabel, i)), SparseTable__TryGetValue_11FD62A8(matrix, (doi_label, i)), SparseTable__TryGetValue_11FD62A8(matrix, (author_list_label, i)), SparseTable__TryGetValue_11FD62A8(matrix, (title_label, i)), SparseTable__TryGetValue_11FD62A8(matrix, (status_label, i)), SparseTable__TryGetValue_11FD62A8(matrix, (status_term_source_reflabel, i)), SparseTable__TryGetValue_11FD62A8(matrix, (status_term_accession_number_label, i)), comments_1)

        return initialize(matrix.ColumnCount, _arrow3461)



def to_sparse_table(publications: FSharpList[Publication]) -> SparseTable:
    matrix: SparseTable = SparseTable_Create_Z2192E64B(None, labels, None, length(publications) + 1)
    comment_keys: FSharpList[str] = empty()
    def action(i: int, p: Publication, publications: Any=publications) -> None:
        i_1: int = (i + 1) or 0
        s_1: dict[str, Any]
        s: OntologyAnnotation = default_arg(p.Status, OntologyAnnotation())
        s_1 = OntologyAnnotation.to_string_object(s, True)
        add_to_dict(matrix.Matrix, (pub_med_idlabel, i_1), default_arg(p.PubMedID, ""))
        add_to_dict(matrix.Matrix, (doi_label, i_1), default_arg(p.DOI, ""))
        add_to_dict(matrix.Matrix, (author_list_label, i_1), default_arg(p.Authors, ""))
        add_to_dict(matrix.Matrix, (title_label, i_1), default_arg(p.Title, ""))
        add_to_dict(matrix.Matrix, (status_label, i_1), s_1["TermName"])
        add_to_dict(matrix.Matrix, (status_term_accession_number_label, i_1), s_1["TermAccessionNumber"])
        add_to_dict(matrix.Matrix, (status_term_source_reflabel, i_1), s_1["TermSourceREF"])
        def f(comment: Comment, i: Any=i, p: Any=p) -> None:
            nonlocal comment_keys
            pattern_input: tuple[str, str] = Comment_toString(comment)
            n: str = pattern_input[0]
            comment_keys = cons(n, comment_keys)
            add_to_dict(matrix.Matrix, (n, i_1), pattern_input[1])

        ResizeArray_iter(f, p.Comments)

    iterate_indexed(action, publications)
    class ObjectExpr3463:
        @property
        def Equals(self) -> Callable[[str, str], bool]:
            def _arrow3462(x: str, y: str) -> bool:
                return x == y

            return _arrow3462

        @property
        def GetHashCode(self) -> Callable[[str], int]:
            return string_hash

    return SparseTable(matrix.Matrix, matrix.Keys, reverse(List_distinct(comment_keys, ObjectExpr3463())), matrix.ColumnCount)


def from_rows(prefix: str | None, line_number: int, rows: IEnumerator[IEnumerable_1[tuple[int, str]]]) -> tuple[str | None, int, FSharpList[Remark], FSharpList[Publication]]:
    tupled_arg: tuple[str | None, int, FSharpList[Remark], SparseTable] = SparseTable_FromRows_Z5579EC29(rows, labels, line_number) if (prefix is None) else SparseTable_FromRows_Z5579EC29(rows, labels, line_number, prefix)
    return (tupled_arg[0], tupled_arg[1], tupled_arg[2], from_sparse_table(tupled_arg[3]))


def to_rows(prefix: str | None, publications: FSharpList[Publication]) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    m: SparseTable = to_sparse_table(publications)
    if prefix is None:
        return SparseTable_ToRows_759CAFC1(m)

    else: 
        return SparseTable_ToRows_759CAFC1(m, prefix)



__all__ = ["pub_med_idlabel", "doi_label", "author_list_label", "title_label", "status_label", "status_term_accession_number_label", "status_term_source_reflabel", "labels", "from_string", "from_sparse_table", "to_sparse_table", "from_rows", "to_rows"]




from __future__ import annotations
from ..dynamic_obj.dynamic_obj import (DynamicObj, DynamicObj_reflection)
from ..fable_library.option import default_arg
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.types import Array
from .inputs import CWLInput
from .outputs import CWLOutput
from .requirements import Requirement

def _expr517() -> TypeInfo:
    return class_type("ARCtrl.CWL.CWLToolDescription", None, CWLToolDescription, DynamicObj_reflection())


class CWLToolDescription(DynamicObj):
    def __init__(self, outputs: Array[CWLOutput], cwl_version: str | None=None, base_command: Array[str] | None=None, requirements: Array[Requirement] | None=None, hints: Array[Requirement] | None=None, inputs: Array[CWLInput] | None=None, metadata: DynamicObj | None=None) -> None:
        super().__init__()
        self._cwlVersion: str = default_arg(cwl_version, "v1.2")
        self._outputs: Array[CWLOutput] = outputs
        self._baseCommand: Array[str] | None = base_command
        self._requirements: Array[Requirement] | None = requirements
        self._hints: Array[Requirement] | None = hints
        self._inputs: Array[CWLInput] | None = inputs
        self._metadata: DynamicObj | None = metadata

    @property
    def CWLVersion(self, __unit: None=None) -> str:
        this: CWLToolDescription = self
        return this._cwlVersion

    @CWLVersion.setter
    def CWLVersion(self, version: str) -> None:
        this: CWLToolDescription = self
        this._cwlVersion = version

    @property
    def Outputs(self, __unit: None=None) -> Array[CWLOutput]:
        this: CWLToolDescription = self
        return this._outputs

    @Outputs.setter
    def Outputs(self, outputs: Array[CWLOutput]) -> None:
        this: CWLToolDescription = self
        this._outputs = outputs

    @property
    def BaseCommand(self, __unit: None=None) -> Array[str] | None:
        this: CWLToolDescription = self
        return this._baseCommand

    @BaseCommand.setter
    def BaseCommand(self, base_command: Array[str] | None=None) -> None:
        this: CWLToolDescription = self
        this._baseCommand = base_command

    @property
    def Requirements(self, __unit: None=None) -> Array[Requirement] | None:
        this: CWLToolDescription = self
        return this._requirements

    @Requirements.setter
    def Requirements(self, requirements: Array[Requirement] | None=None) -> None:
        this: CWLToolDescription = self
        this._requirements = requirements

    @property
    def Hints(self, __unit: None=None) -> Array[Requirement] | None:
        this: CWLToolDescription = self
        return this._hints

    @Hints.setter
    def Hints(self, hints: Array[Requirement] | None=None) -> None:
        this: CWLToolDescription = self
        this._hints = hints

    @property
    def Inputs(self, __unit: None=None) -> Array[CWLInput] | None:
        this: CWLToolDescription = self
        return this._inputs

    @Inputs.setter
    def Inputs(self, inputs: Array[CWLInput] | None=None) -> None:
        this: CWLToolDescription = self
        this._inputs = inputs

    @property
    def Metadata(self, __unit: None=None) -> DynamicObj | None:
        this: CWLToolDescription = self
        return this._metadata

    @Metadata.setter
    def Metadata(self, metadata: DynamicObj | None=None) -> None:
        this: CWLToolDescription = self
        this._metadata = metadata


CWLToolDescription_reflection = _expr517

def CWLToolDescription__ctor_Z3224DB85(outputs: Array[CWLOutput], cwl_version: str | None=None, base_command: Array[str] | None=None, requirements: Array[Requirement] | None=None, hints: Array[Requirement] | None=None, inputs: Array[CWLInput] | None=None, metadata: DynamicObj | None=None) -> CWLToolDescription:
    return CWLToolDescription(outputs, cwl_version, base_command, requirements, hints, inputs, metadata)


__all__ = ["CWLToolDescription_reflection"]




from collections.abc import Callable
from typing import Any
from ...fable_library.util import to_enumerable
from .collections_ import (Dictionary_ofSeq, Dictionary_tryFind)

def OntobeeParser(tsr: str, local_tan: str) -> str:
    return ((((("" + "http://purl.obolibrary.org/obo/") + "") + tsr) + "_") + local_tan) + ""


def BioregistryParser(tsr: str, local_tan: str) -> str:
    return ((((("" + "https://bioregistry.io/") + "") + tsr) + ":") + local_tan) + ""


def OntobeeDPBOParser(tsr: str, local_tan: str) -> str:
    return ((((("" + "http://purl.org/nfdi4plants/ontology/dpbo/") + "") + tsr) + "_") + local_tan) + ""


def MSParser(tsr: str, local_tan: str) -> str:
    return ((((("" + "https://www.ebi.ac.uk/ols4/ontologies/ms/classes/http%253A%252F%252Fpurl.obolibrary.org%252Fobo%252F") + "") + tsr) + "_") + local_tan) + ""


def POParser(tsr: str, local_tan: str) -> str:
    return ((((("" + "https://www.ebi.ac.uk/ols4/ontologies/po/classes/http%253A%252F%252Fpurl.obolibrary.org%252Fobo%252F") + "") + tsr) + "_") + local_tan) + ""


def ROParser(tsr: str, local_tan: str) -> str:
    return ((((("" + "https://www.ebi.ac.uk/ols4/ontologies/ro/classes/http%253A%252F%252Fpurl.obolibrary.org%252Fobo%252F") + "") + tsr) + "_") + local_tan) + ""


def _arrow608(tsr: str) -> Callable[[str], str]:
    def _arrow607(local_tan: str) -> str:
        return OntobeeDPBOParser(tsr, local_tan)

    return _arrow607


def _arrow610(tsr_1: str) -> Callable[[str], str]:
    def _arrow609(local_tan_1: str) -> str:
        return MSParser(tsr_1, local_tan_1)

    return _arrow609


def _arrow612(tsr_2: str) -> Callable[[str], str]:
    def _arrow611(local_tan_2: str) -> str:
        return POParser(tsr_2, local_tan_2)

    return _arrow611


def _arrow614(tsr_3: str) -> Callable[[str], str]:
    def _arrow613(local_tan_3: str) -> str:
        return ROParser(tsr_3, local_tan_3)

    return _arrow613


def _arrow616(tsr_4: str) -> Callable[[str], str]:
    def _arrow615(local_tan_4: str) -> str:
        return BioregistryParser(tsr_4, local_tan_4)

    return _arrow615


def _arrow618(tsr_5: str) -> Callable[[str], str]:
    def _arrow617(local_tan_5: str) -> str:
        return BioregistryParser(tsr_5, local_tan_5)

    return _arrow617


def _arrow620(tsr_6: str) -> Callable[[str], str]:
    def _arrow619(local_tan_6: str) -> str:
        return BioregistryParser(tsr_6, local_tan_6)

    return _arrow619


def _arrow622(tsr_7: str) -> Callable[[str], str]:
    def _arrow621(local_tan_7: str) -> str:
        return BioregistryParser(tsr_7, local_tan_7)

    return _arrow621


def _arrow624(tsr_8: str) -> Callable[[str], str]:
    def _arrow623(local_tan_8: str) -> str:
        return BioregistryParser(tsr_8, local_tan_8)

    return _arrow623


def _arrow626(tsr_9: str) -> Callable[[str], str]:
    def _arrow625(local_tan_9: str) -> str:
        return BioregistryParser(tsr_9, local_tan_9)

    return _arrow625


def _arrow628(tsr_10: str) -> Callable[[str], str]:
    def _arrow627(local_tan_10: str) -> str:
        return BioregistryParser(tsr_10, local_tan_10)

    return _arrow627


def _arrow630(tsr_11: str) -> Callable[[str], str]:
    def _arrow629(local_tan_11: str) -> str:
        return BioregistryParser(tsr_11, local_tan_11)

    return _arrow629


def _arrow632(tsr_12: str) -> Callable[[str], str]:
    def _arrow631(local_tan_12: str) -> str:
        return BioregistryParser(tsr_12, local_tan_12)

    return _arrow631


def _arrow634(tsr_13: str) -> Callable[[str], str]:
    def _arrow633(local_tan_13: str) -> str:
        return BioregistryParser(tsr_13, local_tan_13)

    return _arrow633


def _arrow636(tsr_14: str) -> Callable[[str], str]:
    def _arrow635(local_tan_14: str) -> str:
        return BioregistryParser(tsr_14, local_tan_14)

    return _arrow635


uri_parser_collection: Any = Dictionary_ofSeq(to_enumerable([("DPBO", _arrow608), ("MS", _arrow610), ("PO", _arrow612), ("RO", _arrow614), ("ENVO", _arrow616), ("CHEBI", _arrow618), ("GO", _arrow620), ("OBI", _arrow622), ("PATO", _arrow624), ("PECO", _arrow626), ("TO", _arrow628), ("UO", _arrow630), ("EFO", _arrow632), ("NCIT", _arrow634), ("OMP", _arrow636)]))

def create_oauri(tsr: str, local_tan: str) -> str:
    match_value: Callable[[str, str], str] | None = Dictionary_tryFind(tsr, uri_parser_collection)
    if match_value is None:
        return OntobeeParser(tsr, local_tan)

    else: 
        return match_value(tsr)(local_tan)



__all__ = ["OntobeeParser", "BioregistryParser", "OntobeeDPBOParser", "MSParser", "POParser", "ROParser", "uri_parser_collection", "create_oauri"]




from __future__ import annotations
from typing import Any
from ...fable_library.date import (min_value, try_parse as try_parse_1)
from ...fable_library.types import FSharpRef

def try_parse(s: str) -> Any | None:
    match_value: tuple[bool, Any]
    out_arg: Any = min_value()
    def _arrow605(__unit: None=None, s: Any=s) -> Any:
        return out_arg

    def _arrow606(v: Any, s: Any=s) -> None:
        nonlocal out_arg
        out_arg = v

    match_value = (try_parse_1(s, FSharpRef(_arrow605, _arrow606)), out_arg)
    if match_value[0]:
        return match_value[1]

    else: 
        return None



__all__ = ["try_parse"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.Process.material_attribute import MaterialAttribute
from ...fable_library.list import (choose, of_array, FSharpList)
from ...fable_library.option import map
from ...fable_library.seq import map as map_1
from ...fable_library.util import IEnumerable_1
from ...thoth_json_core.decode import (object, IOptionalGetter, IGetters)
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from ..encode import try_include
from ..idtable import encode
from ..ontology_annotation import (OntologyAnnotation_ROCrate_genID, OntologyAnnotation_ISAJson_encoder, OntologyAnnotation_ISAJson_decoder)

__A_ = TypeVar("__A_")

def gen_id(m: MaterialAttribute) -> str:
    match_value: OntologyAnnotation | None = m.CharacteristicType
    if match_value is None:
        return "#EmptyFactor"

    else: 
        return ("#MaterialAttribute/" + OntologyAnnotation_ROCrate_genID(match_value)) + ""



def encoder(id_map: Any | None, value: MaterialAttribute) -> IEncodable:
    def f(value_1: MaterialAttribute, id_map: Any=id_map, value: Any=value) -> IEncodable:
        def chooser(tupled_arg: tuple[str, IEncodable | None], value_1: Any=value_1) -> tuple[str, IEncodable] | None:
            def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
                return (tupled_arg[0], v_1)

            return map(mapping, tupled_arg[1])

        def _arrow2457(value_2: str, value_1: Any=value_1) -> IEncodable:
            class ObjectExpr2456(IEncodable):
                def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                    return helpers.encode_string(value_2)

            return ObjectExpr2456()

        def _arrow2458(oa: OntologyAnnotation, value_1: Any=value_1) -> IEncodable:
            return OntologyAnnotation_ISAJson_encoder(id_map, oa)

        values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("@id", _arrow2457, gen_id(value_1)), try_include("characteristicType", _arrow2458, value_1.CharacteristicType)]))
        class ObjectExpr2459(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any], value_1: Any=value_1) -> Any:
                def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                    return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_1))

                arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
                return helpers_1.encode_object(arg)

        return ObjectExpr2459()

    if id_map is not None:
        def _arrow2460(m_1: MaterialAttribute, id_map: Any=id_map, value: Any=value) -> str:
            return gen_id(m_1)

        return encode(_arrow2460, f, value, id_map)

    else: 
        return f(value)



def _arrow2462(get: IGetters) -> MaterialAttribute:
    def _arrow2461(__unit: None=None) -> OntologyAnnotation | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("characteristicType", OntologyAnnotation_ISAJson_decoder)

    return MaterialAttribute(None, _arrow2461())


decoder: Decoder_1[MaterialAttribute] = object(_arrow2462)

__all__ = ["gen_id", "encoder", "decoder"]

